import sys
sys.path.append("../")

import os
import numpy as np
import pandas as pd
import torch
from torch import optim

import waterGPT
from waterGPT.utils.save_info import SaveInfo

from waterGPT.utils.data.dataset import Dataset
from waterGPT.utils.data.dataloader import DataLoader
from waterGPT.utils.losses.rmse import RMSELoss
from waterGPT.core.train_attention import TrainEpoch, ValidEpoch, prediction
from waterGPT.config.config import default_config

from waterGPT.utils.data.timefeatures import time_features
from waterGPT.utils.data.camels2 import CAMELS
from waterGPT.utils.data.camels2 import CAMELS_Normalization
from waterGPT.utils.data.load_camels_multiple import LoadCamelsMultiple
from waterGPT.model.TST import Transformer

from waterGPT.utils.stats.metrics import statError

output_path = "./output/"

data_path = "./CAMELS_data"
train_date_list = ["1999-10-01", "2008-09-30"]
valid_date_list = ["1989-10-01", "1999-09-30"]

# Define settings dictionary with default values
settings = {
    "BATCH_SIZE": 256,
    "RHO": 361,
    "DEVICE": "cuda",
    "CUDA_ID": 0,
    "EPOCH": 335,
    "SEED": 6666666,
    "DROPOUT": 0.25,
    "HIDDENSIZE": 512,
    "dim_feedforward": 512,
    "BATCH_FIRST": True,
    "LR": 1e-4,
    "conv_kernel": 2,
    "activation": "gelu",
    "H": 16,
    "N_LAYERS": 4,
    "Eval": False,
    "embConstruct": 1,
    "multiforcing": True,
}


def update_setting(arg):  # use command prompt to update "settings"
    arg = arg.lstrip('-')  # Remove any leading hyphens
    setting_name, new_value = arg.split("=")
    if setting_name in settings:
        current_value = settings[setting_name]
        if isinstance(current_value, bool):
            settings[setting_name] = new_value.lower() in ['true', '1', 'yes']
        else:
            settings[setting_name] = type(current_value)(new_value)


# Overwrite settings with command-line arguments
for arg in sys.argv[1:]:
    update_setting(arg)

assert (settings["HIDDENSIZE"] // settings["H"]) * settings["H"] == settings["HIDDENSIZE"]

torch.cuda.set_device(settings["CUDA_ID"])
device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

var_times = []
var_const = []
target = []

# save info
config_dict = default_config.update(var_times=var_times, var_const=var_const, target=target,
                                    train_date_list=train_date_list, valid_date_list=valid_date_list,
                                    csv_path=data_path, output_path=output_path,
                                    BATCH_SIZE=settings["BATCH_SIZE"], RHO=settings["RHO"],
                                    EPOCH=settings["EPOCH"], DEVICE=settings["DEVICE"],
                                    DROPOUT=settings["DROPOUT"], HIDDENSIZE=settings["HIDDENSIZE"],
                                    dim_feedforward=settings["dim_feedforward"], SEED=settings["SEED"],
                                    conv_kernel=settings["conv_kernel"],
                                    H=settings["H"], N_LAYERS=settings["N_LAYERS"], LR=settings["LR"])
info_saver = SaveInfo()
config_dict = info_saver.update_config(**config_dict)
output_path = config_dict["output_path"]

# load data
if settings["multiforcing"]:
    print("Using multiple forcing data")
    select_id_list = pd.read_csv(os.path.join(os.path.split(waterGPT.__file__)[0], "datasets", "531_basin_list.txt"),
                                 header=None).values[:, 0].tolist()
    train_x, train_y, valid_x, valid_y, obs, stat_dict, var_times, var_const = LoadCamelsMultiple().load_all(
        inp_path=data_path, opt_path=output_path, train_date_list=train_date_list, valid_date_list=valid_date_list,
        select_id_list=select_id_list)
else:
    print("Using single forcing data")
    load_camels = CAMELS(inp_path=data_path, opt_path=output_path)
    train_x, train_y, stat_dict = load_camels.load_train(train_date_list=train_date_list, convert_time_series=True)
    valid_x, valid_y, obs = load_camels.load_test(test_date_list=valid_date_list, stat_dict=stat_dict,
                                                  convert_time_series=True)

de_norm_fn = CAMELS_Normalization(opt_path=output_path).de_normalization
stat_dict["y"] = {"runoff": stat_dict["runoff"]}

# get time features
date_range_train = pd.date_range(train_date_list[0], train_date_list[-1])
date_range_valid = pd.date_range(valid_date_list[0], valid_date_list[-1])
time_stamp_train = time_features(pd.to_datetime(date_range_train), freq="1D")
time_stamp_train = time_stamp_train.transpose(1, 0)
time_stamp_valid = time_features(pd.to_datetime(date_range_valid), freq="1D")
time_stamp_valid = time_stamp_valid.transpose(1, 0)

# set dataset
train_dataset = Dataset(train_x, train_y, batch_size=settings["BATCH_SIZE"],
                        rho=settings["RHO"], stage_name="train", batch_first=settings["BATCH_FIRST"])
valid_dataset = Dataset(valid_x, valid_y, batch_size=settings["BATCH_SIZE"],
                        rho=settings["RHO"], stage_name="valid", batch_first=settings["BATCH_FIRST"])

# set dataloader
train_loader = DataLoader(train_dataset)
valid_loader = DataLoader(valid_dataset)

# set model and loss function
d_input = train_x.shape[-1]
d_output = train_y.shape[-1]
d_model = settings["HIDDENSIZE"]

model = Transformer(c_in=d_input, c_out=d_output, seq_len=settings["RHO"], n_layers=settings["N_LAYERS"],
                    d_model=d_model, n_heads=settings["H"], d_k=None,
                    d_v=None, d_ff=settings["dim_feedforward"], dropout=settings["DROPOUT"],
                    act=settings["activation"], conv_kernel=settings["conv_kernel"],
                    fc_dropout=settings["DROPOUT"], y_range=None, verbose=False,
                    embConstruct=config_dict["embConstruct"], use_batchnorm=config_dict["use_batchnorm"])

loss_fn = RMSELoss()
optim = optim.Adam(model.parameters(), lr=settings["LR"])
monitoring_metrics = []

# set trainer
trainer = TrainEpoch(model, loss_fn, monitoring_metrics, optim, settings["DEVICE"], True,
                     stat_dict=stat_dict, de_norm_fn=de_norm_fn)
valider = ValidEpoch(model, loss_fn, monitoring_metrics, settings["DEVICE"], True,
                     stat_dict=stat_dict, de_norm_fn=de_norm_fn)

best_score = 0
nepoch = settings["EPOCH"]
if not settings["Eval"]:
    for epoch in range(nepoch):
        print("\nEpoch {}".format(epoch))
        train_logs = trainer.run(train_loader, time_stamp_train, train_x, train_y,
                                 rho=settings["RHO"], batch_size=settings["BATCH_SIZE"], stat_dict=stat_dict)
        saveFreq = 20
        if epoch == nepoch - 1:
            saveFreq = nepoch  # make sure it saves the last one
        best_score = info_saver.save_best_model(model, None, epoch, best_score, score_class="NSE_np", save_each=False,
                                                nth_save_model=saveFreq)

# load your model
# model_file = os.path.join(info_saver.save_path, f"model/model_Ep{nepoch-1}.pt")
# model = info_saver.load_model(model_file=model_file)

valider = ValidEpoch(model, loss_fn, monitoring_metrics, settings["DEVICE"], True,
                     stat_dict=stat_dict, de_norm_fn=de_norm_fn)
valid_logs = valider.run(valid_loader, time_stamp_valid, valid_x, valid_y,
                         rho=settings["RHO"], batch_size=settings["BATCH_SIZE"])

de_pred_y = valider.pred
de_obsv_y = valider.obs

np.save(os.path.join(info_saver.save_path, "pred.npy"), de_pred_y)
np.save(os.path.join(info_saver.save_path, "obs.npy"), de_obsv_y)

metrics_dict = statError(pred=de_pred_y[:, :, 0], target=de_obsv_y[:, :, 0])
print("KGE: {}".format(np.nanmedian(metrics_dict["KGE12"])))
print("Corr: {}".format(np.nanmedian(metrics_dict["Corr"])))
print("RMSE: {}".format(np.nanmedian(metrics_dict["RMSE"])))
