import sys
import torch
import numpy as np
from tqdm import tqdm as tqdm
from .meter import AverageValueMeter
from waterGPT.utils.stats.transform import convert_to_numpy, align_dim
from waterGPT.utils.stats.transform import normalization


class Epoch:
    def __init__(self, model, loss, metrics, stage_name, device="cpu", verbose=True, stat_dict=None, de_norm_fn=None):
        self.model = model
        self.loss = loss
        self.metrics = metrics
        self.stage_name = stage_name
        self.verbose = verbose
        self.device = device
        self.stat_dict = stat_dict
        self.de_norm_fn = de_norm_fn

        self._to_device()

    def _to_device(self):
        self.model.to(self.device)
        self.loss.to(self.device)
        for metric in self.metrics:
            metric.to(self.device)

    def _format_logs(self, logs):
        str_logs = ["{} : {:.3}".format(k, v) for k, v in logs.items()]
        s = ", ".join(str_logs)
        return s

    def batch_update(self, x, y):
        raise NotImplementedError

    def on_epoch_start(self):
        pass

    def run(self, dataloader, *args, **kwargs):

        self.on_epoch_start()

        logs = {}
        loss_meter = AverageValueMeter()
        metrics_meters = {metric.__name__: AverageValueMeter() for metric in self.metrics}

        with tqdm(
                dataloader,
                desc=self.stage_name,
                file=sys.stdout,
                disable=not (self.verbose),
                ncols=2000,
                bar_format='{desc:<10}{percentage:3.0f}%|{bar:20}{r_bar}'

        ) as iterator:
            pred_list, obs_list = [], []
            for ndx, (x, y, slice_grid, slice_time) in enumerate(iterator):
                ndx_list = slice_grid.tolist()
                x, y = x.to(self.device), y.to(self.device)  # basin, time, feas
                loss, y_pred = self.batch_update(x, y)

                # update loss logs
                loss_value = convert_to_numpy(loss)
                loss_meter.add(loss_value)
                # loss_logs = {self.loss.__name__: loss_meter.mean}
                loss_logs = {"loss": loss_meter.mean}
                logs.update(loss_logs)

                # update metrics logs
                # de-normalization for calculating metrics
                stat_dict = list(self.stat_dict["y"].values())
                y_pred_de, y_de = de_normaliza_data(pred=y_pred, obs=y, stat_list=stat_dict, de_norm_fn=self.de_norm_fn,
                                                    ndx_list=ndx_list)
                metrics_logs = {
                    metric_fn.__name__: float(convert_to_numpy(metric_fn(y_pred_de, y_de))) for
                    metric_fn in self.metrics}

                # update metrics logs
                logs.update(metrics_logs)

                if self.verbose:
                    s = self._format_logs(logs)
                    iterator.set_postfix_str(s)

                # record prediction
                if not self.stage_name == "train":
                    pred_list.append(convert_to_numpy(y_pred))
                    obs_list.append(convert_to_numpy(y))

            # calculate actually validation metrics
            basin_dim = 0
            if not self.stage_name == "train":
                pred_list = np.concatenate(pred_list, axis=basin_dim)
                obs_list = np.concatenate(obs_list, axis=basin_dim)
                stat_dict = list(self.stat_dict["y"].values())
                pred_list_de, obs_list_de = de_normaliza_data(pred=pred_list, obs=obs_list, stat_list=stat_dict,
                                                              de_norm_fn=self.de_norm_fn)
                metrics_logs = {metric_fn.__name__: float(convert_to_numpy(metric_fn(pred_list_de, obs_list_de))) for
                                metric_fn in self.metrics}
                iterator.write('Validation metrics: {}'.format(self._format_logs(metrics_logs)))
                # update metrics logs
                logs.update(metrics_logs)
        return logs

    def prediction(self):
        pass

class TrainEpoch(Epoch):
    def __init__(self, model, loss, metrics, optimizer, device="cpu", verbose=True, stat_dict=None, DI=None,
                 de_norm_fn=None, *args, **kwargs):
        super().__init__(
            model=model,
            loss=loss,
            metrics=metrics,
            stage_name="train",
            device=device,
            verbose=verbose,
            stat_dict=stat_dict,
            de_norm_fn=de_norm_fn,
        )
        self.optimizer = optimizer
        self.DI = DI

    def on_epoch_start(self):
        self.model.train()

    def batch_update(self, x, y):
        self.optimizer.zero_grad()

        if self.DI is None:
            prediction = self.model.forward(x)
        else:
            prediction = run_with_DI(self.model, x)

        loss = self.loss(prediction, y)
        loss.backward()
        self.optimizer.step()
        return loss, prediction


class ValidEpoch(Epoch):
    def __init__(self, model, loss, metrics, device="cpu", verbose=True, stat_dict=None, DI=None, de_norm_fn=None, *args, **kwargs):
        super().__init__(
            model=model,
            loss=loss,
            metrics=metrics,
            stage_name="valid",
            device=device,
            verbose=verbose,
            stat_dict=stat_dict,
            de_norm_fn=de_norm_fn,
        )
        self.DI = DI

    def on_epoch_start(self):
        self.model.eval()

    def batch_update(self, x, y):
        self.model.eval()
        with torch.no_grad():

            if self.DI is None:
                prediction = self.model.forward(x)
            else:
                prediction = run_with_DI(self.model, x)

            loss = self.loss(prediction, y)
        return loss, prediction


def run_with_DI(model, x):
    index_nan = torch.isnan(x[:, :, -1])
    x[:, :, -1][index_nan] = 0.0
    prediction = model.forward(x)
    x_clone = torch.clone(x)
    x_clone[:, :, -1][index_nan] = prediction[:, :, -1][index_nan]
    prediction = model.forward(x_clone)
    return prediction


def prediction(model, x, obs, stat_dict, metrics, batch_first=True, DI=None, de_norm_fn=None, batch_size=200):
    """
    x: basin, time, features

    """
    model.eval()
    with torch.no_grad():
        num_pixels = len(x)
        pred_list = []
        x_split_list = np.split(x, np.arange(0, num_pixels, batch_size), axis=0)[1:]
        for x_sub in x_split_list:
            if not batch_first:
                x_sub = np.transpose(x_sub, [1, 0, 2])  # time, basin, features
            x_torch = torch.from_numpy(x_sub.astype(np.float32)).cuda()
            # print(next(model.parameters()).device)
            if not DI is None:
                pred_torch = run_with_DI(model, x_torch)
            else:
                pred_torch = model(x_torch)
            pred = convert_to_numpy(pred_torch)
            if not batch_first:
                pred = np.transpose(pred, [1, 0, 2])  # basin, time, features
            pred_list.append(pred)
        pred_list = np.concatenate(pred_list, axis=0)

        stat_dict_test = list(stat_dict["y"].values())
        pred_list_de, obs_de = de_normaliza_data(pred=pred_list, obs=obs, stat_list=stat_dict_test,
                                                 de_norm_fn=de_norm_fn)

        if (obs is None) or len(obs) == 0:
            logs = None
        else:
            logs = {metric_fn.__name__: float(metric_fn(pred_list_de, obs_de)) for metric_fn in metrics}
        # logs = ", ".join(["{} - {:.5}".format(k, v) for k, v in logs.items()])

    return pred_list_de, logs


def de_normaliza_data(pred, stat_list, obs=None, de_norm_fn=None, **kwargs):
    if de_norm_fn is None:
        de_norm_fn = normalization
    pred = de_norm_fn(pred, stat_list=stat_list, from_raw=False, kwargs=kwargs)
    if not obs is None:
        obs = de_norm_fn(obs, stat_list=stat_list, from_raw=False, kwargs=kwargs)
    return pred, obs
