from waterGPT.core.train import Epoch, TrainEpoch, ValidEpoch
import sys
import torch
import numpy as np
from tqdm import tqdm as tqdm
from .meter import AverageValueMeter
from waterGPT.utils.stats.transform import convert_to_numpy, align_dim
from waterGPT.utils.stats.transform import normalization
import copy


def run_with_DI(model, x_enc, x_enc_mark, stage_name):
    x = x_enc
    index_nan = torch.isnan(x[:, :, -1])
    x[:, :, -1][index_nan] = 0.0
    prediction = model.forward(x_enc, x_enc_mark, stage_name=stage_name)
    x_clone = torch.clone(x)
    x_clone[:, :, -1][index_nan] = prediction[:, :, -1][index_nan]
    prediction = model.forward(x_clone, x_enc_mark, stage_name=stage_name)
    return prediction


def adjust_learning_rate(optimizer, epoch, learning_rate, method="fixed"):
    assert method == "fixed"

    assert method in ["epoch", "fixed"]
    # lr = args.learning_rate * (0.2 ** (epoch // 2))
    if method in ["epoch"]:
        lr_adjust = {epoch: learning_rate * (0.99 ** ((epoch - 1) // 1))}
    elif method in ["fixed"]:
        lr_adjust = {
            50: 5e-2,
            100: 1e-2,
            150: 5e-3,
            200: 1e-3,
            250: 5e-4,
        }
    if epoch in lr_adjust.keys():
        lr = lr_adjust[epoch]
        for param_group in optimizer.param_groups:
            param_group['lr'] = lr
        print('Updating learning rate to {}'.format(lr))


def slice_index(num_grid, num_time, batch_size, rho):
    """
    data size: grid, time, features
    """
    num_grid_split = num_grid // batch_size
    num_time_split = num_time // rho
    index_grid = np.arange(0, num_grid)
    index_time = np.arange(0, num_time)
    index_grid_list = np.array_split(index_grid, num_grid_split)
    index_time_list = np.array_split(index_time, num_time_split)
    return index_grid_list, index_time_list


def prediction(self, model, x, obs, stat_dict, metrics, batch_first=True, DI=None, de_norm_fn=None, batch_size=200,
               rho=360, time_features=None, device="cuda"):
    """
    x: basin, time, features

    """

    num_grid, num_time, num_dim = x.shape
    num_grid_full = int(np.ceil(num_grid / batch_size)) * batch_size
    num_time_full = int(np.ceil(num_time / rho)) * rho

    time_features_copy = copy.deepcopy(time_features)
    time_features = np.repeat(time_features_copy[None, :, :], num_grid, axis=0)

    if num_grid != num_grid_full:
        x = torch.concat([x, x[:num_grid_full - num_grid, :, :]], axis=0)
        obs = torch.concat([obs, obs[:num_grid_full - num_grid, :, :]], axis=0)
        time_features = np.concatenate([time_features, time_features[:num_grid_full - num_grid, :, :]], axis=0)

    if num_time != num_time_full:
        x = torch.concat([x, x[:, :num_time_full - num_time, :]], axis=1)
        obs = torch.concat([obs, obs[:, :num_time_full - num_time, :]], axis=1)
        time_features = np.concatenate([time_features, time_features[:, :num_time_full - num_time, :]], axis=1)

    assert batch_size

    # num_grid, num_time, _ = x.shape
    index_grid_list, index_time_list = slice_index(num_grid=num_grid_full, num_time=num_time_full,
                                                   batch_size=batch_size, rho=rho)

    pred_y = np.ones_like(obs)

    model.eval()
    with torch.no_grad():

        for idx_grid, slice_grid in enumerate(index_grid_list):
            for idx_time, slice_time in enumerate(index_time_list):

                x_sub = x[slice_grid[:, None], slice_time, :]  # basin, time, features

                # get time features
                sub_time = time_features[slice_grid[:, None], slice_time, :]

                # to the device
                if type(x_sub) in [np.float32, np.float16, np.float64, np.ndarray]:
                    x_torch = torch.from_numpy(x_sub.astype(np.float32)).to(device)
                else:
                    x_torch = x_sub.to(device)
                sub_time = torch.from_numpy(sub_time).float().to(device)

                # input the model
                x_enc = x_torch
                x_enc_mark = sub_time

                if DI is None:
                    pred_torch = model(x_enc, x_enc_mark, stage_name="valid")
                else:
                    pred_torch = run_with_DI(model, x_enc, x_enc_mark, stage_name="valid")

                pred = convert_to_numpy(pred_torch)
                pred_y[slice_grid[:, None], slice_time, :] = pred

        # calculate loss and metrics
        pred_list = pred_y

        pred_list = pred_list[:num_grid, : num_time, :]
        obs = obs[:num_grid, : num_time, :]

        loss = self.loss(torch.from_numpy(pred_list), obs)

        # stat_dict_test = list(stat_dict["y"].values())
        try:
            stat_dict_test = stat_dict[0]
        except:
            stat_dict_test = stat_dict
        # stat_dict_test = stat_dict
        pred_list_de, obs_de = de_normaliza_data(pred=pred_list, obs=obs, stat_list=stat_dict_test,
                                                 de_norm_fn=de_norm_fn)

        self.pred = pred_list_de
        self.obs = obs_de

        if (obs is None) or len(obs) == 0:
            logs = None
        else:
            logs = {metric_fn.__name__: float(metric_fn(pred_list_de[:,:,0], obs_de[:,:,0])) for metric_fn in metrics}

        loss_logs = {"loss": float(loss.numpy())}
        logs.update(loss_logs)

    return loss, logs


def run(self, dataloader, time_features=None, x_all=None, y_all=None, batch_size=200, rho=360, pred_len=None, stat_dict=None, *args,
        **kwargs):
    self.on_epoch_start()

    logs = {}
    loss_meter = AverageValueMeter()
    metrics_meters = {metric.__name__: AverageValueMeter() for metric in self.metrics}

    time_features_copy = copy.deepcopy(time_features)
    if stat_dict is None:
        stat_dict = list(self.stat_dict["y"].values())

    with tqdm(
            dataloader,
            desc=self.stage_name,
            file=sys.stdout,
            disable=not (self.verbose),
            ncols=2000,
            bar_format='{desc:<10}{percentage:3.0f}%|{bar:20}{r_bar}'

    ) as iterator:
        if self.stage_name in ["train"]:

            for ndx, (x, y, slice_grid, slice_time) in enumerate(iterator):

                num_pixels = len(slice_grid)
                time_features = np.repeat(time_features_copy[None, :, :], num_pixels, axis=0)
                if pred_len is None:
                    # DO NOT REMOVE np.arange(num_pixels)
                    sub_time = time_features[np.arange(num_pixels)[:, None], np.array(slice_time), :]
                    sub_time = torch.from_numpy(sub_time).float()

                    ndx_list = slice_grid.tolist()
                    x, y, sub_time = x.to(self.device), y.to(self.device), sub_time.to(self.device)  # basin, time, feas
                    loss, y_pred = self.batch_update(x, y, sub_time)
                else:
                    sub_time_x = time_features[np.arange(num_pixels)[:, None], np.array(slice_time[0]), :]
                    sub_time_y = time_features[np.arange(num_pixels)[:, None], np.array(slice_time[1]), :]
                    ndx_list = slice_grid.tolist()
                    # basin, time, feas
                    x, y, sub_time_x, sub_time_y = x.to(self.device), y.to(self.device), sub_time_x.to(
                        self.device), sub_time_y.to(self.device)
                    loss, y_pred = self.batch_update(x, y, sub_time_x, sub_time_y)

                # update loss logs
                loss_value = convert_to_numpy(loss)
                loss_meter.add(loss_value)
                # loss_logs = {self.loss.__name__: loss_meter.mean}
                loss_logs = {"loss": loss_meter.mean}
                logs.update(loss_logs)

                # update metrics logs
                # de-normalization for calculating metrics
                y_pred_de, y_de = de_normaliza_data(pred=y_pred, obs=y, stat_list=stat_dict, de_norm_fn=self.de_norm_fn,
                                                    ndx_list=ndx_list)
                metrics_logs = {
                    metric_fn.__name__: float(convert_to_numpy(metric_fn(y_pred_de[:,:,0], y_de[:,:,0]))) for
                    metric_fn in self.metrics}

                # update metrics logs
                logs.update(metrics_logs)

                if self.verbose:
                    s = self._format_logs(logs)
                    iterator.set_postfix_str(s)

        x = x_all
        y = y_all
        x = torch.from_numpy(x).float()
        y = torch.from_numpy(y).float()

        loss, logs = self.prediction(model=self.model, x=x, obs=y, stat_dict=stat_dict, metrics=self.metrics,
                                     de_norm_fn=self.de_norm_fn, time_features=time_features_copy,
                                     batch_size=batch_size, rho=rho, DI=self.DI)

    return logs


# modify run function in Epoch
Epoch.run = run
Epoch.prediction = prediction


class TrainEpoch(TrainEpoch):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        try:
            self.lr_scheduler = kwargs["lr_scheduler"]
        except:
            self.lr_scheduler = None

    def batch_update(self, x, y, sub_time, sub_time_y=None, *args, **kwargs):
        self.optimizer.zero_grad()

        if sub_time_y is None:
            x_enc = x
            x_enc_mark = sub_time

            if self.DI is None:
                prediction = self.model.forward(x_enc, x_enc_mark, stage_name="train")
            else:
                prediction = run_with_DI(self.model, x_enc, x_enc_mark, stage_name="train")
        else:
            x_enc = x
            x_enc_mark = sub_time
            y_enc = y
            y_enc_mark = sub_time_y
            prediction = self.model.forward(x_enc, x_enc_mark, y_enc=y_enc, y_enc_mark=y_enc_mark, stage_name="train")

        loss = self.loss(prediction, y)
        loss.backward()
        self.optimizer.step()

        if self.lr_scheduler is not None:
            # print("lr: ", self.lr_scheduler.get_lr())
            self.lr_scheduler.step()
            self.optimizer.zero_grad()

        return loss, prediction


class ValidEpoch(ValidEpoch):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def batch_update(self, x, y, sub_time, sub_time_y=None, *args, **kwargs):
        self.model.eval()
        with torch.no_grad():

            if sub_time_y is None:
                x_enc = x
                x_enc_mark = sub_time

                if self.DI is None:
                    prediction = self.model.forward(x_enc, x_enc_mark, stage_name="valid")
                else:
                    prediction = run_with_DI(self.model, x_enc, x_enc_mark, stage_name="valid")
            else:
                x_enc = x
                x_enc_mark = sub_time
                y_enc = y
                y_enc_mark = sub_time_y
                prediction = self.model.forward(x_enc, x_enc_mark, y_enc=y_enc, y_enc_mark=y_enc_mark,stage_name="valid")
            # prediction = self.model.forward(x, sub_time)
            loss = self.loss(prediction, y)
        return loss, prediction


def de_normaliza_data(pred, stat_list, obs=None, de_norm_fn=None, **kwargs):
    if de_norm_fn is None:
        de_norm_fn = normalization
    pred = de_norm_fn(pred, stat_list=stat_list, from_raw=False, kwargs=kwargs)
    if not obs is None:
        obs = de_norm_fn(obs, stat_list=stat_list, from_raw=False, kwargs=kwargs)
    return pred, obs
