import math

import torch
from torch import nn
from torch.nn import Module
from torch.nn import functional as F
from torch.nn import Parameter

from waterGPT.model.dropout import DropMask, createMask
from waterGPT.utils.data.dataset_attention_embed import DataEmbedding
from waterGPT.utils.data.CausalConv1d import CausalConv1d, CausalConv1dAlt, CausalConv1dAlt2

class IdentityNet(nn.Module):
    def __init__(self):
        super(IdentityNet, self).__init__()

    def forward(self, x):
        return x

class Transpose(Module):
    def __init__(self, *dims, contiguous=False):
        super().__init__()
        self.dims, self.contiguous = dims, contiguous

    def forward(self, x):
        if self.contiguous:
            return x.transpose(*self.dims).contiguous()
        else:
            return x.transpose(*self.dims)

    def __repr__(self):
        if self.contiguous:
            return f"{self.__class__.__name__}(dims={', '.join([str(d) for d in self.dims])}).contiguous()"
        else:
            return f"{self.__class__.__name__}({', '.join([str(d) for d in self.dims])})"



class _ScaledDotProductAttention(Module):
    def __init__(self, d_k):
        super().__init__()
        self.d_k = d_k
        self.identity = IdentityNet()

    def forward(self, q, k, v, mask=None):
        """
        Parameters
            q: [bs, n_heads, seq_len, d_k]
            k: [bs, n_heads, d_k, seq_len]
            v: [bs, n_heads, seq_len, d_v]

        Returns
            context: [bs, n_heads, q_len, d_v]
            attention: [bs, n_heads, q_len, q_len]
        """
        q, k, v = self.identity(q), self.identity(k), self.identity(v)
        # matmul (q, k) - similarity scores for all pairs of positions in an input sequence
        scores = torch.matmul(q, k)  # [bs, n_heads, seq_len, seq_len]

        # Scale
        scores = scores / (self.d_k ** 0.5)

        # mask (optional)
        if mask is not None:
            scores.masked_fill_(mask, -1e9)

        # softMax
        attn = F.softmax(scores, dim=-1)  # [bs, n_heads, q_len, q_len]

        context = torch.matmul(attn, v)  # [bs, n_heads, q_len, d_v]

        return context, attn


class _MultiHeadAttention(Module):
    def __init__(self, d_model, n_heads, d_k, d_v, dropout=0.5):

        """
        Parameters
            d_model: dimension of the features
            n_heads: number of heads
            d_k: dimension of the key
            d_v: dimension of the value
            dropout: dropout rate
        """
        super().__init__()
        self.dropout = 0.5
        self.d_model = d_model
        self.n_heads, self.d_k, self.d_v = n_heads, d_k, d_v

        # self.W_Q = nn.Linear(d_model, d_k * n_heads, bias=False)
        # self.W_K = nn.Linear(d_model, d_k * n_heads, bias=False)
        # self.W_V = nn.Linear(d_model, d_v * n_heads, bias=False)

        self.W_O = nn.Linear(n_heads * d_v, d_model, bias=False)

        self.w_qh = Parameter(torch.Tensor(d_model, d_k * n_heads))
        self.w_kh = Parameter(torch.Tensor(d_model, d_k * n_heads))
        self.w_vh = Parameter(torch.Tensor(d_model, d_k * n_heads))
        # self.w_oh = Parameter(torch.Tensor(n_heads * d_v, d_model))

        self.b_q = None
        self.b_k = None
        self.b_v = None
        # self.b_o = None

        self.reset_parameters()

    def reset_parameters(self):

        initrange = 1.0 / math.sqrt(self.d_model)
        # list(self.parameters())[3].shape
        for weight in self.parameters():
            weight.data.uniform_(-initrange, initrange)

    def reset_mask(self, x):
        self.mask_w_qh = createMask(self.w_qh, self.dropout)
        self.mask_w_kh = createMask(self.w_kh, self.dropout)
        self.mask_w_vh = createMask(self.w_vh, self.dropout)
        # self.mask_w_oh = createMask(self.w_oh, self.dropout)

    def forward(self, Q, K, V, mask=None, stage_name=None):
        """
        Parameters
            Q, K, V: [bs x seq_len x d_model]
            mask: [q_len x q_len]

        Returns
            output: [bs, seq_len, d_model]
            attn: [bs, n_heads, seq_len, seq_len]
        """

        assert stage_name in ["train", "valid"]
        bs = Q.size(0)

        if stage_name == "train":
            do_drop_mask = True
        else:
            do_drop_mask = False

        if do_drop_mask:
            # reset mask
            self.reset_mask(Q)

            # do drop mask to reduce the overfitting
            w_qh = DropMask.apply(self.w_qh, self.mask_w_qh, True)
            w_kh = DropMask.apply(self.w_kh, self.mask_w_kh, True)
            w_vh = DropMask.apply(self.w_vh, self.mask_w_vh, True)
            # w_oh = DropMask.apply(self.w_oh, self.mask_w_oh, True)
        else:
            w_qh = self.w_qh
            w_kh = self.w_kh
            w_vh = self.w_vh
            # w_oh = self.w_oh

        # get initial Q, K, V
        # q_s: [bs, n_heads, seq_len, d_k]
        q_s = F.linear(Q, w_qh, self.b_q).view(bs, -1, self.n_heads, self.d_k).transpose(1, 2)
        # k_s: [bs, n_heads, d_k, seq_len]
        k_s = F.linear(K, w_kh, self.b_k).view(bs, -1, self.n_heads, self.d_k).permute(0, 2, 3, 1)
        # k_s: [bs, n_heads, seq_len, d_v]
        v_s = F.linear(V, w_vh, self.b_v).view(bs, -1, self.n_heads, self.d_v).transpose(1, 2)

        # scaled dot-product attention (multiple heads)
        # context: [bs, n_heads, q_len, d_v], attn: [bs, n_heads, q_len, q_len]
        context, attn = _ScaledDotProductAttention(self.d_k)(q_s, k_s, v_s)

        # concat. context: [bs, q_len, n_heads * d_v]
        context = context.transpose(1, 2).contiguous().view(bs, -1, self.n_heads * self.d_v)

        # lLinear. context: [bs, q_len, d_model]
        output = self.W_O(context)

        return output, attn


def get_activation_fn(activation):
    if activation == "relu":
        return nn.ReLU()
    elif activation == "gelu":
        return nn.GELU()
    else:
        return activation()


class _TSTEncoderLayer(Module):
    def __init__(self, q_len, d_model, n_heads, d_k=None, d_v=None, d_ff=256, dropout=0.1, activation="gelu", conv_kernel=2):
        """
        Parameters
            q_len: length of the time sequence
            d_model: dimension of the features
            n_heads: number of heads
            d_k: dimension of the key
            d_v: dimension of the value
            d_ff: dimension of the feed-forward layer
            dropout: dropout rate
            activation: activation function
        """
        super().__init__()

        assert d_model // n_heads, f"d_model ({d_model}) must be divisible by n_heads ({n_heads})"
        d_k = d_model // n_heads if d_k is None else d_k
        d_v = d_model // n_heads if d_v is None else d_v

        # multi-head attention
        self.self_attn = _MultiHeadAttention(d_model, n_heads, d_k, d_v, dropout=dropout)

        # add residual connection with residual dropout & batch norm
        self.dropout_attn = nn.Dropout(dropout)
        self.batchnorm_attn = nn.Sequential(
            Transpose(1, 2),
            nn.BatchNorm1d(d_model),
            Transpose(1, 2)
        )

        # position-wise feed-forward
        self.ff = nn.Sequential(
            nn.Linear(d_model, d_ff),
            get_activation_fn(activation),
            nn.Dropout(dropout),
            nn.Linear(d_ff, d_model)
        )

        # add residual connection with residual dropout & batch norm
        self.dropout_ffn = nn.Dropout(dropout)
        self.batchnorm_ffn = nn.Sequential(
            Transpose(1, 2),
            nn.BatchNorm1d(d_model),
            Transpose(1, 2)
        )
        if conv_kernel>0:
            self.conv = CausalConv1dAlt(in_channels=d_model, out_channels=d_model, kernel_width=conv_kernel)

    def forward(self, src, mask=None, stage_name=None):
        """
        Parameters
            src: mini-batch of input sequences [bs, seq_len, d_model]
            stage_name: "train", "valid", or "test"

        Returns
            output: [bs, seq_len, d_model]
        """
        assert stage_name in ["train", "valid"]

        # multi-Head attention sublayer
        # src2: [bs, seq_len, d_model], attn: [bs, n_heads, seq_len, seq_len]
        #src = self.conv(src)
        src2, attn = self.self_attn(src, src, src, mask=mask, stage_name=stage_name)

        # add residual connection with residual dropout & batch norm
        src = src + self.dropout_attn(src2)  # [bs, seq_len, d_model]
        src = self.batchnorm_attn(src)

        # position-wise feed-forward sublayer
        src2 = self.ff(src)  # [bs, seq_len, d_model]

        # add residual connection with residual dropout & batch norm
        src = src + self.dropout_ffn(src2)  # [bs, seq_len, d_model]
        src = self.batchnorm_ffn(src)

        return src


class _TSTEncoder(Module):
    def __init__(self, q_len, d_model, n_heads, d_k=None, d_v=None, d_ff=None, dropout=0.1, activation='gelu',
                 n_layers=1):
        """
        Multi Layer Encoder
        """
        super().__init__()

        self.layers = nn.ModuleList(
            [_TSTEncoderLayer(q_len, d_model, n_heads=n_heads, d_k=d_k, d_v=d_v, d_ff=d_ff, dropout=dropout,
                              activation=activation) for i in range(n_layers)]
        )

    def forward(self, src, stage_name=None):
        """
        Parameters
            src: mini-batch of input sequences [bs, seq_len, d_model]
            stage_name: "train", "valid", or "test"

        Returns
            output: [bs, seq_len, d_model]
        """
        assert stage_name in ["train", "valid"]
        output = src
        for mod in self.layers:
            output = mod(output, stage_name=stage_name)  # [bs, seq_len, d_model]
        return output


class Transformer(Module):
    def __init__(self, c_in, c_out, seq_len, n_layers=4, d_model=256, n_heads=4, d_k=None, d_v=None, d_ff=512,
                 dropout=0.05, act="gelu", conv_kernel=0, embConstruct=0, use_batchnorm=False, *args, **kwargs):
        """
        Parameters
            c_in: number of input channels
            c_out: number of output channels
            seq_len: length of the time sequence
            n_layers: number of encoder layers
            d_model: dimension of the features
            n_heads: number of heads
            d_k: dimension of the key
            d_v: dimension of the value
            d_ff: dimension of the feed-forward layer
            dropout: dropout rate
            act: activation function
        """
        super().__init__()

        # input encoding
        q_len = seq_len

        # encoder
        self.encoder = _TSTEncoder(q_len, d_model, n_heads, d_k=d_k, d_v=d_v, d_ff=d_ff, dropout=dropout,
                                   activation=act, n_layers=n_layers)

        # head
        self.head_nf = q_len * d_model
        self.d_model = d_model
        self.project = self.create_head(c_out, act=act)

        self.enc_embedding = DataEmbedding(c_in=c_in, d_model=d_model, dropout=dropout, conv_kernel=conv_kernel, embConstruct=embConstruct, use_batchnorm=use_batchnorm)
        # self.decoder = nn.Linear(self.d_model, c_out)

    def create_head(self, c_out, act="gelu"):
        head = nn.Sequential(
            get_activation_fn(act),
            nn.Linear(self.d_model, c_out)
        )

        return head

    def forward(self, x, time_mark, stage_name=None, *args, **kwargs):
        """
        Parameters
            x: input time series: [bs, q_len, n_vars]
            time_mark: time mark of each time step: [bs, q_len, 3(dayOfWeek, dayOfMonth, monthOfYear)]
            stage_name: "train", "valid" or "test"

        Returns
            output: [bs, q_len, c_out]
        """
        assert stage_name in ["train", "valid"]

        # data embedding + time embedding + position embedding
        x_embed = self.enc_embedding(x, time_mark)  # [bs, q_len, d_model]

        # encoder
        x_enc = self.encoder(x_embed, stage_name=stage_name)  # [bs, q_len, d_model]

        output = self.project(x_enc)  # [bs, q_len, c_out]
        return output
