import math
import torch
import torch.nn as nn
class CausalConv1d(nn.Module): # 256_res_256...
    def __init__(self, in_channels, out_channels, kernel_width, num_layers=2, dilation=1, weights=None, use_batchnorm=False, apply_final_activation=False):
        super(CausalConv1d, self).__init__()
        self.num_layers = num_layers
        self.use_batchnorm = use_batchnorm
        self.apply_final_activation = apply_final_activation
        self.convs = nn.ModuleList()
        self.bns = nn.ModuleList() if use_batchnorm else None
        self.relu = nn.ReLU()

        # Calculate channel sizes for each layer
        mid_channels = out_channels // 1
        channel_sizes = [mid_channels if i < num_layers - 1 else out_channels for i in range(num_layers)]

        # Create all layers
        for i in range(num_layers):
            layer_in_channels = in_channels if i == 0 else channel_sizes[i - 1]
            layer_out_channels = channel_sizes[i]
            layer = nn.Conv1d(layer_in_channels, layer_out_channels, kernel_width,
                              padding=(kernel_width - 1) * dilation, dilation=dilation)
            self.init_weights(layer, weights if i == 0 else None)
            self.convs.append(layer)

            if use_batchnorm:
                bn = nn.BatchNorm1d(layer_out_channels)
                self.bns.append(bn)

    def init_weights(self, layer, weights=None):
        if weights is not None:
            layer.weight.data = torch.tensor(weights).view(layer.out_channels, layer.in_channels, layer.kernel_size[0])
            layer.bias.data.zero_()
        else:
            # Initialize weights using kaiming_uniform_ (He initialization)
            nn.init.kaiming_uniform_(layer.weight, a=math.sqrt(5))

            # Initialize biases using a uniform distribution
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(layer.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(layer.bias, -bound, bound)

    def forward(self, x):
        x = x.transpose(1, 2)
        for i, (conv, bn) in enumerate(zip(self.convs, self.bns) if self.use_batchnorm else zip(self.convs, [None] * len(self.convs))):
            if i > 0:  # Save the input of the second last layer for the residual connection
                residual = x
            x = conv(x)
            x = x[:, :, :-(conv.kernel_size[0] - 1) * conv.dilation[0]]
            if self.use_batchnorm:
                x = bn(x)
            if i < self.num_layers - 1 or self.apply_final_activation:
                x = self.relu(x)
            #if i == self.num_layers - 2:  # Apply the residual connection after the second last layer
            if i > 0:  # Apply the residual connection after the second last layer
                x = x + residual
        x = x.transpose(1, 2)
        return x


class CausalConv1dAlt(nn.Module):  # 128_res_128,256
    def __init__(self, in_channels, out_channels, kernel_width, num_layers=3, dilation=1, weights=None, use_batchnorm=False, apply_final_activation=False):
        super(CausalConv1dAlt, self).__init__()
        self.num_layers = num_layers
        self.use_batchnorm = use_batchnorm
        self.apply_final_activation = apply_final_activation
        self.convs = nn.ModuleList()
        self.bns = nn.ModuleList() if use_batchnorm else None
        self.relu = nn.ReLU()

        # Calculate channel sizes for each layer
        mid_channels = out_channels // 2
        channel_sizes = [mid_channels if i < num_layers - 1 else out_channels for i in range(num_layers)]

        # Create all layers
        for i in range(num_layers):
            layer_in_channels = in_channels if i == 0 else channel_sizes[i - 1]
            layer_out_channels = channel_sizes[i]
            layer = nn.Conv1d(layer_in_channels, layer_out_channels, kernel_width,
                              padding=(kernel_width - 1) * dilation, dilation=dilation)
            self.init_weights(layer, weights if i == 0 else None)
            self.convs.append(layer)

            if use_batchnorm:
                bn = nn.BatchNorm1d(layer_out_channels)
                self.bns.append(bn)

    def init_weights(self, layer, weights=None):
        if weights is not None:
            layer.weight.data = torch.tensor(weights).view(layer.out_channels, layer.in_channels, layer.kernel_size[0])
            layer.bias.data.zero_()
        else:
            # Initialize weights using kaiming_uniform_ (He initialization)
            nn.init.kaiming_uniform_(layer.weight, a=math.sqrt(5))

            # Initialize biases using a uniform distribution
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(layer.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(layer.bias, -bound, bound)

    def forward(self, x):
        x = x.transpose(1, 2)
        for i, (conv, bn) in enumerate(zip(self.convs, self.bns) if self.use_batchnorm else zip(self.convs, [None] * len(self.convs))):
            if i == self.num_layers - 2:  # Save the input of the second last layer for the residual connection
                residual = x
            x = conv(x)
            x = x[:, :, :-(conv.kernel_size[0] - 1) * conv.dilation[0]]
            if self.use_batchnorm:
                x = bn(x)
            if i < self.num_layers - 1 or self.apply_final_activation:
                x = self.relu(x)
            if i == self.num_layers - 2:  # Apply the residual connection after the second last layer
                x = x + residual
        x = x.transpose(1, 2)
        return x

class CausalConv1dAlt2(nn.Module): # 128,128,256 last layer fc
    def __init__(self, in_channels, out_channels, kernel_width, num_layers=3, dilation=1, weights=None, use_batchnorm=False, apply_final_activation=False):
        super(CausalConv1dAlt2, self).__init__()
        self.num_layers = num_layers
        self.use_batchnorm = use_batchnorm
        self.apply_final_activation = apply_final_activation
        self.convs = nn.ModuleList()
        self.bns = nn.ModuleList() if use_batchnorm else None
        self.relu = nn.ReLU()

        # Calculate channel sizes for each layer
        mid_channels = out_channels // 2
        channel_sizes = [mid_channels if i < num_layers - 1 else out_channels for i in range(num_layers)]

        # Create all layers
        for i in range(num_layers - 1):
            layer_in_channels = in_channels if i == 0 else channel_sizes[i - 1]
            layer_out_channels = channel_sizes[i]
            layer = nn.Conv1d(layer_in_channels, layer_out_channels, kernel_width,
                              padding=(kernel_width - 1) * dilation, dilation=dilation)
            self.init_weights(layer, weights if i == 0 else None)
            self.convs.append(layer)

            if use_batchnorm:
                bn = nn.BatchNorm1d(layer_out_channels)
                self.bns.append(bn)

        # Add the fully connected layer as the last layer
        self.fc = nn.Linear(mid_channels, out_channels)

    def init_weights(self, layer, weights=None):
        if weights is not None:
            layer.weight.data = torch.tensor(weights).view(layer.out_channels, layer.in_channels, layer.kernel_size[0])
            layer.bias.data.zero_()
        else:
            # Initialize weights using kaiming_uniform_ (He initialization)
            nn.init.kaiming_uniform_(layer.weight, a=math.sqrt(5))

            # Initialize biases using a uniform distribution
            fan_in, _ = nn.init._calculate_fan_in_and_fan_out(layer.weight)
            bound = 1 / math.sqrt(fan_in)
            nn.init.uniform_(layer.bias, -bound, bound)

    def forward(self, x):
        x = x.transpose(1, 2)
        for i, (conv, bn) in enumerate(zip(self.convs, self.bns) if self.use_batchnorm else zip(self.convs, [None] * len(self.convs))):
            if i == self.num_layers - 2:  # Save the input of the second last layer for the residual connection
                residual = x
            x = conv(x)
            x = x[:, :, :-(conv.kernel_size[0] - 1) * conv.dilation[0]]
            if self.use_batchnorm:
                x = bn(x)
            x = self.relu(x)
            if i == self.num_layers - 2:  # Apply the residual connection after the second last layer
                x = x + residual
        x = x.transpose(1, 2)

        # Apply the fully connected layer
        x = self.fc(x)

        if self.apply_final_activation:
            x = self.relu(x)

        return x