import os
import json
import warnings
import zipfile
import numpy as np
import pandas as pd
import urllib.request
from waterGPT.utils.sys_tools import re_folder
from waterGPT.utils.stats.transform import normalization

class CAMELS():

    def __init__(self, inp_path=None, opt_path=None, select_date_range=None,
                 forcing_list=None, constant_list=None, target_list=None, dataset="NLDAS",
                 select_id_list=None, all_date_range=None, attribute_txt_list=None, info_name="camles_info.json"):

        # # detect CAMELS files
        # self.download_camles(opt_path=inp_path)

        if opt_path is None:
            opt_path = "./"

        if select_date_range is None:
            select_date_range = ["1980-01-01", "2014-12-31"]

        if all_date_range is None:
            all_date_range = ["1980-01-01", "2014-12-31"]

        if forcing_list is None:
            forcing_list = ['dayl', 'prcp', 'srad', 'tmax', 'tmin', 'vp']

        if constant_list is None:
            constant_list = ['elev_mean', 'slope_mean', 'area_gages2', 'frac_forest', 'lai_max', 'lai_diff',
                             'dom_land_cover_frac', 'dom_land_cover', 'root_depth_50', 'soil_depth_statsgo',
                             'soil_porosity', 'soil_conductivity', 'max_water_content', 'geol_1st_class',
                             'geol_2nd_class', 'geol_porostiy', 'geol_permeability']

        if attribute_txt_list is None:
            attribute_txt_list = ['topo', 'clim', 'hydro', 'vege', 'soil', 'geol']

        if target_list is None:
            target_list = ["runoff"]

        self.inp_path = inp_path
        self.opt_path = opt_path
        self.select_date_range = select_date_range
        self.all_date_range = all_date_range
        self.forcing_list = forcing_list
        self.constant_list = constant_list
        self.attribute_txt_list = attribute_txt_list
        self.target_list = target_list
        self.dataset = dataset
        self.num_all_time = len(pd.date_range(all_date_range[0], all_date_range[-1]))

        try:
            # load basin info
            self.normalize_tools = CAMELS_Normalization()
            self.df_info = self.get_info(inp_path=inp_path, select_id_list=select_id_list)
            self.basin_area, self.basin_prec = self.load_basin_norm_info(self.inp_path, self.df_info)
            # save basic info
            info_dict = {
                "huc": self.df_info["huc"].values.tolist(),
                "id": self.df_info["id"].values.tolist(),
                "name": self.df_info["name"].values.tolist(),
                "lat": self.df_info["lat"].values.tolist(),
                "lon": self.df_info["lon"].values.tolist(),
                "area": self.df_info["area"].values.tolist(),
                "basin_area": self.basin_area, "basin_prec": self.basin_prec, }
            with open(os.path.join(opt_path, info_name), "w") as fout:
                json.dump(info_dict, fout, indent=4)
        except:
            pass

    def load_basin_norm_info(self, inp_path, df_info):
        huc_list = df_info["huc"].values.tolist()
        select_id_list = df_info["id"].values.tolist()

        tmp1 = os.path.join(inp_path, "camels_attributes_v2.0")
        inp_area = os.path.join(tmp1, "camels_topo.txt")
        inp_prec = os.path.join(tmp1, "camels_clim.txt")
        df_area = pd.read_csv(inp_area, sep=';')
        df_prec = pd.read_csv(inp_prec, sep=';')

        # select by id list
        df_area_sel = self.select_by_id(df_area, select_id_list)
        df_prec_sel = self.select_by_id(df_prec, select_id_list)

        return df_area_sel["area_gages2"].values.tolist(), df_prec_sel["p_mean"].values.tolist()

    def download_camles(self, opt_path=None, download_extended=False):

        if opt_path is None:
            opt_path = self.inp_path

        forcing_path = os.path.join(opt_path, "basin_timeseries_v1p2_metForcing_obsFlow")
        attribute_path = os.path.join(opt_path, "camels_attributes_v2.0/")
        re_folder(forcing_path)
        re_folder(attribute_path)

        # download forcing data
        base_url = "https://gdex.ucar.edu/dataset/camels/file/"

        attribute_txt_list = ["camels_clim", "camels_geol", "camels_hydro", "camels_hydro", "camels_name",
                              "camels_soil", "camels_topo", "camels_vege"]
        attribute_url_list = [os.path.join(base_url, x + ".txt") for x in attribute_txt_list]
        attribute_url = os.path.join(base_url, "camels_attributes_v2.0.xlsx")
        forcing_url = os.path.join(base_url, "basin_timeseries_v1p2_metForcing_obsFlow.zip")

        print("downloading forcing ...")
        forcing_file = os.path.join(forcing_path, "basin_timeseries_v1p2_metForcing_obsFlow.zip")
        # if not os.path.exists(forcing_file):
        urllib.request.urlretrieve(forcing_url, forcing_file)
        print("downloading attributes ...")
        urllib.request.urlretrieve(attribute_url, os.path.join(attribute_path, "camels_attributes_v2.0.xlsx"))
        for dw_url in attribute_url_list:
            filename = os.path.split(dw_url)[-1]
            urllib.request.urlretrieve(dw_url, os.path.join(attribute_path, filename))
        print("done!")

        print("unzip forcing ...")
        with zipfile.ZipFile(os.path.join(forcing_path, "basin_timeseries_v1p2_metForcing_obsFlow.zip"),
                             'r') as zip_ref:
            zip_ref.extractall(forcing_path)
        print("done!")

        if download_extended:
            import tarfile
            extended_path = os.path.join(forcing_path, "basin_dataset_public_v1p2/basin_mean_forcing")

            print("download maurer extended and unzip ...")
            maurer_extended_url = "https://www.hydroshare.org/resource/17c896843cf940339c3c3496d0c1c077/data/contents/maurer_extended.zip"
            maurer_extended_file = os.path.join(extended_path, "maurer_extended.zip")
            urllib.request.urlretrieve(maurer_extended_url, maurer_extended_file)
            with zipfile.ZipFile(maurer_extended_file, 'r') as zip_ref: zip_ref.extractall(extended_path)
            print("done !")

            print("download nldas extended and unzip ...")
            nldas_extended_url = "https://www.hydroshare.org/resource/0a68bfd7ddf642a8be9041d60f40868c/data/contents/nldas_extended.tar.xz"
            nldas_extended_file = os.path.join(extended_path, "nldas_extended.tar.xz")
            urllib.request.urlretrieve(nldas_extended_url, nldas_extended_file)
            # unzip
            file = tarfile.open(nldas_extended_file)
            file.extractall(extended_path)
            file.close()
            print("done")

    def load_train(self, train_date_list=None, convert_time_series=True, simple_norm=False):

        # load data
        print("loading training data ...")
        x, y, c = self.load_camles(inp_path=self.inp_path, select_date_range=train_date_list,
                                   forcing_list=self.forcing_list, constant_list=self.constant_list,
                                   target_list=self.target_list, dataset=self.dataset,
                                   attribute_txt_list=self.attribute_txt_list, re_extract=True)

        if simple_norm:
            # calculate statistics
            stat_dict_x = {}
            for ndx, var_name in enumerate(self.forcing_list):
                stat_dict_x[var_name] = self.normalize_tools.cal_stat(x[:, :, ndx])
            stat_dict_y = {}
            for ndx, var_name in enumerate(self.target_list):
                stat_dict_y[var_name] = self.normalize_tools.cal_stat(y[:, :, ndx])
            stat_dict_c = {}
            for ndx, var_name in enumerate(self.constant_list):
                stat_dict_c[var_name] = self.normalize_tools.cal_stat(c[:, ndx])
            stat_dict = {**stat_dict_x, **stat_dict_y, **stat_dict_c}
            with open(os.path.join(self.opt_path, "stat_dict.json"), "w") as fout:
                json.dump(stat_dict, fout, indent=4)

            # normalization
            x_norm = normalization(data=x, stat_list=[stat_dict[x] for x in self.forcing_list], from_raw=True)
            y_norm = normalization(data=y, stat_list=[stat_dict[x] for x in self.target_list], from_raw=True)
            c_norm = normalization(data=c, stat_list=[stat_dict[x] for x in self.constant_list], from_raw=True)
        else:
            # do basin normalization for runoff
            y_basin_norm = self.normalize_tools.basin_norm(y, area_list=self.basin_area, prec_list=self.basin_prec,
                                                           from_raw=True, ndx_list=None)
            # calculate runoff std for each basin
            y_basin_std = []
            for ndx_basin in range(len(y_basin_norm)):
                y_basin_std.append(CAMELS_Normalization().cal_stat_gamma(y_basin_norm[ndx_basin])[-1])
            y_basin_std = np.array(y_basin_std)
            with open(os.path.join(self.opt_path, "y_basin_std.json"), "w") as fout:
                json.dump(y_basin_std.tolist(), fout, indent=4)

            # calculate statistics
            stat_dict = self.normalize_tools.cal_xyc_stat(x=x, x_name_list=self.forcing_list, y=y_basin_norm,
                                                          y_name_list=self.target_list, c=c, c_name_list=self.constant_list)
            with open(os.path.join(self.opt_path, "stat_dict.json"), "w") as fout:
                json.dump(stat_dict, fout, indent=4)

            # normalization
            x_norm = self.normalize_tools.norm_by_dict(x=x, varLst=self.forcing_list, staDic=stat_dict, toNorm=True)
            y_norm = self.normalize_tools.norm_by_dict(x=y_basin_norm, varLst=self.target_list, staDic=stat_dict,
                                                       toNorm=True)
            c_norm = self.normalize_tools.norm_by_dict(x=c, varLst=self.constant_list, staDic=stat_dict, toNorm=True)

        # remove nan
        x_norm[np.isnan(x_norm)] = 0.0
        c_norm[np.isnan(c_norm)] = 0.0

        if convert_time_series:
            # repeat time dimension
            c_norm = c_norm[:, None, :]  # basin, time, features
            c_norm = np.repeat(c_norm, x_norm.shape[1], axis=1)
            # concat x and c
            x_norm = np.concatenate([x_norm, c_norm], axis=-1)

            return x_norm, y_norm, stat_dict
        else:
            return x_norm, c_norm, y_norm, stat_dict

    def load_test(self, test_date_list=None, stat_dict=None, stat_dict_path=None, convert_time_series=True, simple_norm=False):

        print("loading testing data ...")
        # if stat_dict_path is None and stat_dict is None:
        #     stat_dict_path = os.path.join(self.opt_path, 'stat_dict.json')

        # if not stat_dict_path is None:
        #     stat_dict = json.load(open(stat_dict_path))

        assert not stat_dict is None

        # load data
        x, y, c = self.load_camles(inp_path=self.inp_path, select_date_range=test_date_list,
                                   forcing_list=self.forcing_list, constant_list=self.constant_list,
                                   target_list=self.target_list, dataset=self.dataset,
                                   attribute_txt_list=self.attribute_txt_list, re_extract=False)
        obs = y

        if simple_norm:
            # normalization
            x_norm = normalization(data=x, stat_list=[stat_dict[x] for x in self.forcing_list], from_raw=True)
            y_norm = normalization(data=y, stat_list=[stat_dict[x] for x in self.target_list], from_raw=True)
            c_norm = normalization(data=c, stat_list=[stat_dict[x] for x in self.constant_list], from_raw=True)
        else:
            # do basin normalization for runoff
            y_basin_norm = self.normalize_tools.basin_norm(y, area_list=self.basin_area, prec_list=self.basin_prec,
                                                           from_raw=True, ndx_list=None)
            # normalization
            x_norm = self.normalize_tools.norm_by_dict(x=x, varLst=self.forcing_list, staDic=stat_dict, toNorm=True)
            y_norm = self.normalize_tools.norm_by_dict(x=y_basin_norm, varLst=self.target_list, staDic=stat_dict,
                                                       toNorm=True)
            c_norm = self.normalize_tools.norm_by_dict(x=c, varLst=self.constant_list, staDic=stat_dict, toNorm=True)

        # remove nan
        x_norm[np.isnan(x_norm)] = 0.0
        c_norm[np.isnan(c_norm)] = 0.0
        print("done!")
        if convert_time_series:
            # repeat time dimension
            c_norm = c_norm[:, None, :]  # basin, time, features
            c_norm = np.repeat(c_norm, x_norm.shape[1], axis=1)
            # concat x and c
            x_norm = np.concatenate([x_norm, c_norm], axis=-1)

            return x_norm, y_norm, obs
        else:
            return x_norm, c_norm, y_norm, obs

    def load_camles(self, inp_path, select_date_range=None, forcing_list=None, constant_list=None, target_list=None,
                    dataset="NLDAS", attribute_txt_list=None, re_extract=True):

        df_info = self.df_info

        # format input path
        dataset = self.format_ds(dataset)

        if re_extract:
            # load x for the whole time period
            self.all_x = self.get_all_forcing(inp_path=inp_path, forcing_list=forcing_list, df_info=df_info,
                                              dataset=dataset)

            # load y for the whole time period
            self.all_y = self.get_all_runoff(inp_path=inp_path, target_list=target_list, df_info=df_info,
                                             dataset=dataset)
            # load c
            self.c = self.get_attribute(inp_path=inp_path, attribute_txt_list=attribute_txt_list, df_info=df_info,
                                        constant_list=constant_list)

        # select time range: assuming all inputs have the same size
        date_range = pd.date_range(self.all_date_range[0], self.all_date_range[-1])
        time_df = pd.Series(index=date_range, data=np.arange(len(date_range)))
        time_index = time_df.loc[pd.date_range(select_date_range[0], select_date_range[-1])].values

        x = self.all_x[:, time_index, :]
        y = self.all_y[:, time_index, :]
        c = self.c

        return x, y, c

    def format_ds(self, dataset):
        assert isinstance(dataset, str)
        ds = dataset.lower()
        assert ds in ["nldas", "daymet", "maurer", "nldas_extended", "maurer_extended"]

        # update dataset folder location
        if ds == "maurer_extended":
            # self.forcing_list = ["dayl", "prcp", "srad", "swe", "tmax", "tmin", "vp"]
            self.txt_ds = "maurer"
        elif ds == "nldas_extended":
            self.txt_ds = "nldas"
        elif ds == "daymet":
            self.txt_ds = "cida"
        else:
            self.txt_ds = ds

        return ds

    def get_attribute(self, inp_path, attribute_txt_list, df_info, constant_list):
        huc_list = df_info["huc"].values.tolist()
        select_id_list = df_info["id"].values.tolist()

        df_list = []
        tmp1 = os.path.join(inp_path, "camels_attributes_v2.0")
        for attribute_name in attribute_txt_list:
            inp_file = os.path.join(tmp1, "camels_{}.txt".format(attribute_name))
            data_df = pd.read_csv(inp_file, sep=';')

            # select by id list
            data_df_sel = self.select_by_id(data_df, select_id_list)

            # drop gage_id
            data_df_sel.pop("gauge_id")

            df_list.append(data_df_sel)
        df_list = pd.concat(df_list, axis=1)  # basin, features
        df_list = df_list[constant_list]

        # factorize
        for column_name in df_list:
            if pd.api.types.is_string_dtype(df_list[column_name]):
                value, ref = pd.factorize(df_list[column_name], sort=True)
                df_list[column_name] = value

        return df_list.values

    def select_by_id(self, data_df, select_id_list):

        data_df = data_df.set_index("gauge_id")
        data_df = data_df.loc[select_id_list]
        data_df = data_df.reset_index()
        return data_df

    def get_all_runoff(self, inp_path, target_list, df_info, dataset="nldas"):
        huc_list = df_info["huc"].values.tolist()
        select_id_list = df_info["id"].values.tolist()
        tmp1 = os.path.join(inp_path,
                            "basin_timeseries_v1p2_metForcing_obsFlow/basin_dataset_public_v1p2/usgs_streamflow")

        data_list = []
        inconsistent_list = []
        for ndx, (huc, gage_id) in enumerate(zip(huc_list, select_id_list)):
            inp_file = os.path.join(tmp1, "{:02d}/{:08d}_streamflow_qc.txt".format(huc, gage_id))
            data_df = pd.read_csv(inp_file, sep=r'\s+', header=None)
            data_df.columns = ["id", "year", "month", "day", "runoff", "qc"]

            if data_df.shape[0] != self.num_all_time:
                data_df = self.fill_time_df(data_df=data_df)
                inconsistent_list.append(gage_id)

            # # select time. It is time consuming, we have to do it because each site have different time period
            # data_df_sel = self.select_df_by_time(data_df, select_date_range)
            data_df_sel = data_df

            # select variables
            runoff = data_df_sel["runoff"].values
            runoff[runoff < 0] = np.nan

            data_list.append(runoff)

        # save inconsistent files name
        if not os.path.exists(os.path.join(self.inp_path, "inconsistent_runoff_ids.json")):
            with open(os.path.join(self.inp_path, "inconsistent_runoff_ids.json"), "w") as fout:
                json.dump(inconsistent_list, fout, indent=4)

        data_list = np.stack(data_list, axis=0)  # basin, time
        data_list = data_list[:, :, None]

        return data_list

    def get_all_forcing(self, inp_path, forcing_list, df_info, dataset="nldas"):
        huc_list = df_info["huc"].values.tolist()
        select_id_list = df_info["id"].values.tolist()

        tmp1 = os.path.join(inp_path,
                            "basin_timeseries_v1p2_metForcing_obsFlow/basin_dataset_public_v1p2/basin_mean_forcing")

        data_list = []
        for ndx, (huc, gage_id) in enumerate(zip(huc_list, select_id_list)):
            inp_file = os.path.join(tmp1, "{}/{:02d}/{:08d}_lump_{}_forcing_leap.txt".format(
                dataset, huc, gage_id, self.txt_ds))
            data_df = pd.read_csv(inp_file, sep=r'\s+', skiprows=3)

            # remove unit
            columns_list = data_df.columns
            columns_list = [x.lower() for x in columns_list]
            columns_list = [x.split("(")[0] for x in columns_list]
            columns_list = [x.replace("mnth", "month") for x in columns_list]
            data_df.columns = columns_list
            assert set(forcing_list) <= set(columns_list)

            if data_df.shape[0] != self.num_all_time:
                data_df = self.fill_time_df(data_df=data_df)

            # if ndx == 0:
            #     first_shape = data_df.shape
            #     assert first_shape[0] == self.num_all_time
            # assert data_df.shape == first_shape

            # # select by time. too slow
            # data_df["date"] = pd.to_datetime(dict(year=data_df.year, month=data_df.mnth, day=data_df.day))
            # data_df_sel = self.select_df_by_time(data_df, select_date_range)
            data_df_sel = data_df

            # select forcing variable
            data_df_sel = data_df_sel[forcing_list]

            data_list.append(data_df_sel.values)
        data_list = np.stack(data_list, axis=0)  # basins, time, features

        return data_list

    def fill_time_df(self, data_df):
        # fill data with nan
        idx = pd.date_range(self.all_date_range[0], self.all_date_range[-1])
        # data_df = data_df.set_index("date")
        data_df.index = pd.to_datetime(dict(year=data_df.year, month=data_df.month, day=data_df.day))
        data_df = data_df.reindex(idx, fill_value=np.nan)
        # data_df = data_df.reset_index().rename(columns={"index": 'date'})
        data_df = data_df.reset_index(drop=True)
        return data_df

    def select_df_by_time(self, data_df, select_date_range, fill_time=False):

        if fill_time:
            data_df = self.fill_time_df(data_df=data_df)

        data_df = data_df.set_index("date")

        date_range = list(pd.date_range(select_date_range[0], select_date_range[1]).strftime("%Y-%m-%d"))
        data_df = data_df.loc[date_range]
        data_df = data_df.reset_index().rename(columns={data_df.index.name: 'date'})
        return data_df

    def get_info(self, inp_path, select_id_list=None):
        info_file = os.path.join(inp_path,
                                 "basin_timeseries_v1p2_metForcing_obsFlow/basin_dataset_public_v1p2/basin_metadata/gauge_information.txt")
        df = pd.read_csv(info_file, sep='\t', header=None, skiprows=1)
        df.columns = ['huc', 'id', 'name', 'lat', 'lon', 'area']

        if not select_id_list is None:
            df.set_index("id", inplace=True)
            df = df.loc[select_id_list]
            df = df.reset_index().rename(columns={df.index.name: 'id'})

        return df


class CAMELS_Normalization():

    def __init__(self, inp_path=None, opt_path=None, info_name=None):
        self.inp_path = inp_path
        self.opt_path = opt_path
        self.info_name = info_name

    def cal_xyc_stat(self, x=None, y=None, c=None, x_name_list=None, y_name_list=None, c_name_list=None):

        if not x is None:
            assert not x_name_list is None
            stat_dict_x = self.cal_statistics(data=x, var_list=x_name_list, is_timeseries=True)
        else:
            stat_dict_x = {}

        if not y is None:
            assert not y_name_list is None
            stat_dict_y = self.cal_statistics(data=y, var_list=y_name_list, is_timeseries=True)
        else:
            stat_dict_y = {}

        if not c is None:
            assert not c_name_list is None
            stat_dict_c = self.cal_statistics(data=c, var_list=c_name_list, is_timeseries=False)

        # concatenate
        stat_dict = {**stat_dict_x, **stat_dict_y, **stat_dict_c}
        return stat_dict

    def basin_norm(self, x, area_list, prec_list, from_raw=True, ndx_list=None):

        nd = len(x.shape)
        # prec_list = np.ones(len(area_list))

        if not ndx_list is None:
            area_list = area_list[ndx_list]
            prec_list = prec_list[ndx_list]

        assert len(x.shape) == 3

        temparea = np.tile(np.array(area_list)[:, None], (1, x.shape[1]))[:, :, None]
        tempprep = np.tile(np.array(prec_list)[:, None], (1, x.shape[1]))[:, :, None]

        if from_raw:
            # (m^3/day)/(m^3/day)
            flow = (x * 0.0283168 * 3600 * 24) / ((temparea * (10 ** 6)) * (tempprep * 10 ** (-3)))
        else:
            flow = x * ((temparea * (10 ** 6)) * (tempprep * 10 ** (-3))) / (0.0283168 * 3600 * 24)

        assert len(flow.shape) == 3
        return flow

    def cal_statistics(self, data, var_list, is_timeseries=True):

        stat_dict = {}
        if is_timeseries:
            for ndx, var_name in enumerate(var_list):
                if var_name in ["prcp", "Precip", "runoff", "Runoff", "Runofferror", "MSWEP", "GPM", ]:
                    # gamma distribution
                    stat_dict[var_name] = self.cal_stat_gamma(data[:, :, ndx])
                else:
                    stat_dict[var_name] = self.cal_stat(data[:, :, ndx])
        else:
            # constant data
            for ndx, var_name in enumerate(var_list):
                stat_dict[var_name] = self.cal_stat(data[:, ndx])

        return stat_dict

    def cal_stat_gamma(self, x):
        """
        for daily streamflow and precipitation
        :param x:
        :return:
        """
        a = x.flatten()
        b = a[~np.isnan(a)]  # remove nan
        # do some tranformation to change gamma characteristics
        b = np.log10(np.sqrt(b) + 0.1)
        p10 = np.percentile(b, 10).astype(float)
        p90 = np.percentile(b, 90).astype(float)
        mean = np.mean(b).astype(float)
        std = np.std(b).astype(float)
        if std < 0.001:
            std = 1
        return [p10, p90, mean, std]

    def cal_stat(self, x):
        a = x.flatten()
        b = a[~np.isnan(a)]  # remove nan
        p10 = np.percentile(b, 10).astype(float)
        p90 = np.percentile(b, 90).astype(float)
        mean = np.mean(b).astype(float)
        std = np.std(b).astype(float)
        if std < 0.001:
            std = 1
        return [p10, p90, mean, std]

    def norm_by_dict(self, x, varLst, staDic, *, toNorm):
        if type(varLst) is str:
            varLst = [varLst]
        out = np.zeros(x.shape)
        for k in range(len(varLst)):
            var = varLst[k]
            stat = staDic[var]
            if toNorm is True:
                if len(x.shape) == 3:
                    if var in ['prcp', 'usgsFlow', 'Precip', 'runoff', 'Runoff', 'Runofferror']:
                        temp = np.log10(np.sqrt(x[:, :, k]) + 0.1)
                        out[:, :, k] = (temp - stat[2]) / stat[3]
                    else:
                        out[:, :, k] = (x[:, :, k] - stat[2]) / stat[3]
                elif len(x.shape) == 2:
                    if var in ['prcp', 'usgsFlow', 'Precip', 'runoff', 'Runoff', 'Runofferror']:
                        temp = np.log10(np.sqrt(x[:, k]) + 0.1)
                        out[:, k] = (temp - stat[2]) / stat[3]
                    else:
                        out[:, k] = (x[:, k] - stat[2]) / stat[3]
            else:
                if len(x.shape) == 3:
                    out[:, :, k] = x[:, :, k] * stat[3] + stat[2]
                    if var in ['prcp', 'usgsFlow', 'Precip', 'runoff', 'Runoff', 'Runofferror']:
                        temptrans = np.power(10, out[:, :, k]) - 0.1
                        # I expect to see RuntimeWarnings in this block
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore", category=RuntimeWarning)
                            temptrans[temptrans < 0] = 0  # set negative as zero
                        out[:, :, k] = (temptrans) ** 2
                elif len(x.shape) == 2:
                    out[:, k] = x[:, k] * stat[3] + stat[2]
                    if var in ['prcp', 'usgsFlow', 'Precip', 'runoff', 'Runoff', 'Runofferror']:
                        temptrans = np.power(10, out[:, k]) - 0.1
                        temptrans[temptrans < 0] = 0
                        out[:, k] = (temptrans) ** 2
        return out

    def de_normalization(self, data, stat_dict=None, stat_dict_path=None, from_raw=False, **kwargs):
        try:
            ndx_list = kwargs["kwargs"]["ndx_list"]
        except:
            ndx_list = None

        try:
            kfold_index = kwargs["kwargs"]["kfold_index"]
            assert ndx_list is None
            ndx_list = kfold_index
        except:
            pass

        data = self.convert_to_numpy(data)

        if stat_dict is None:
            if stat_dict_path is None and stat_dict is None:
                stat_dict_path = os.path.join(self.opt_path, 'stat_dict.json')

        if not stat_dict_path is None:
            stat_dict = json.load(open(stat_dict_path))

        assert not stat_dict is None

        if type(stat_dict) is list:
            stat_dict = {"runoff": stat_dict[0]}

        # transform back to the original observation
        data = self.transNormbyDic(data, 'runoff', stat_dict, toNorm=False)
        data = self.basinNorm(data, "All", toNorm=False, ndx_list=ndx_list)

        # change the units ft3/s to m3/s
        data = data * 0.0283168
        return data

    def transNormbyDic(self, x, varLst, staDic, *, toNorm):
        if type(varLst) is str:
            varLst = [varLst]
        out = np.zeros(x.shape)
        for k in range(len(varLst)):
            var = varLst[k]
            stat = staDic[var]
            if toNorm is True:
                if len(x.shape) == 3:
                    if var in ['prcp', 'usgsFlow', 'Precip', 'runoff', 'Runoff', 'Runofferror']:
                        temp = np.log10(np.sqrt(x[:, :, k]) + 0.1)
                        out[:, :, k] = (temp - stat[2]) / stat[3]
                    else:
                        out[:, :, k] = (x[:, :, k] - stat[2]) / stat[3]
                elif len(x.shape) == 2:
                    if var in ['prcp', 'usgsFlow', 'Precip', 'runoff', 'Runoff', 'Runofferror']:
                        temp = np.log10(np.sqrt(x[:, k]) + 0.1)
                        out[:, k] = (temp - stat[2]) / stat[3]
                    else:
                        out[:, k] = (x[:, k] - stat[2]) / stat[3]
            else:
                if len(x.shape) == 3:
                    out[:, :, k] = x[:, :, k] * stat[3] + stat[2]
                    if var in ['prcp', 'usgsFlow', 'Precip', 'runoff', 'Runoff', 'Runofferror']:
                        temptrans = np.power(10, out[:, :, k]) - 0.1
                        # I expect to see RuntimeWarnings in this block
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore", category=RuntimeWarning)
                            temptrans[temptrans < 0] = 0  # set negative as zero
                        out[:, :, k] = (temptrans) ** 2
                elif len(x.shape) == 2:
                    out[:, k] = x[:, k] * stat[3] + stat[2]
                    if var in ['prcp', 'usgsFlow', 'Precip', 'runoff', 'Runoff', 'Runofferror']:
                        temptrans = np.power(10, out[:, k]) - 0.1
                        temptrans[temptrans < 0] = 0
                        out[:, k] = (temptrans) ** 2
        return out

    def basinNorm(self, x, gageid, toNorm, ndx_list=None):

        if self.info_name == None:
            info_name = "camles_info.json"
        else:
            info_name = self.info_name

        # load info
        with open(os.path.join(self.opt_path, info_name), "r") as read_file:
            info_dict = json.load(read_file)
        gageid = info_dict["id"]
        basinarea = np.array(info_dict["basin_area"])[:, None]
        meanprep = np.array(info_dict["basin_prec"])[:, None]
        # meanprep = np.ones(len(basinarea))[:, None]
        # for regional training, gageid should be numpyarray
        nd = len(x.shape)

        if not ndx_list is None:
            basinarea = basinarea[ndx_list]
            meanprep = meanprep[ndx_list]

        # meanprep = readAttr(gageid, ['q_mean'])
        if nd == 3 and x.shape[2] == 1:
            x = x[:, :, 0]  # unsqueeze the original 3 dimension matrix
        temparea = np.tile(basinarea, (1, x.shape[1]))
        tempprep = np.tile(meanprep, (1, x.shape[1]))
        if toNorm is True:
            flow = (x * 0.0283168 * 3600 * 24) / (
                    (temparea * (10 ** 6)) * (tempprep * 10 ** (-3)))  # (m^3/day)/(m^3/day)
        else:

            flow = x * ((temparea * (10 ** 6)) * (tempprep * 10 ** (-3))) / (0.0283168 * 3600 * 24)
        if nd == 3:
            flow = np.expand_dims(flow, axis=2)
        return flow

    def convert_to_numpy(self, data):
        import torch
        data_type = type(data)
        if data_type == torch.Tensor:
            return data.cpu().detach().numpy()
        else:
            return data

