"""
code from:
https://github.com/thuml/autoformer

"""
import torch
import torch.nn as nn
from waterGPT.utils.data.CausalConv1d import CausalConv1d, CausalConv1dAlt, CausalConv1dAlt2

import math


class PositionalEmbedding(nn.Module):
    """
    create a new parameter position enconding in nn.Module, but do not participate in the gradient update (optimizer.step())
    https://zhuanlan.zhihu.com/p/338592312
    """

    def __init__(self, d_model, max_len=5000):  # max_len >= seq_len/ time_len
        super(PositionalEmbedding, self).__init__()
        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model).float()  # shape = (5000, 512), max_len similar to time_len
        pe.require_grad = False

        position = torch.arange(0, max_len).float().unsqueeze(1)  # shape = (5000, 1)
        div_term = (torch.arange(0, d_model, 2).float() * -(math.log(10000.0) / d_model)).exp()  # shape = (256)

        pe[:, 0::2] = torch.sin(position * div_term)  # shape = (5000, 256), even location
        pe[:, 1::2] = torch.cos(position * div_term)  # odd location

        pe = pe.unsqueeze(0)  # shape = (1, 5000, 512)
        self.register_buffer('pe', pe)

    def forward(self, x):
        return self.pe[:, :x.size(1)]


class LinearEmbedding(nn.Module):
    def __init__(self, c_in, d_model, initrange=None):
        super(LinearEmbedding, self).__init__()
        self.initrange = initrange
        self.c_in = c_in
        self.d_model = d_model

        self.linear = nn.Linear(c_in, d_model, bias=False)

        # initial parameters.
        self.reset_parameters()

    def reset_parameters(self):
        if self.initrange is None:
            initrange = 1.0 / math.sqrt(self.d_model)
        else:
            initrange = self.initrange

        for weight in self.parameters():
            weight.data.uniform_(-initrange, initrange)

    def forward(self, x):
        x = self.linear(x)
        return x

class Time2Vec(nn.Module):
    def __init__(self, inp_dim, opt_dim):
        super(Time2Vec, self).__init__()
        self.opt_dim = opt_dim
        self.inp_dim = inp_dim
        self.w0 = nn.parameter.Parameter(torch.randn(inp_dim, 1))
        self.b0 = nn.parameter.Parameter(torch.randn(1))
        self.w = nn.parameter.Parameter(torch.randn(inp_dim, opt_dim - 1))
        self.b = nn.parameter.Parameter(torch.randn(opt_dim - 1))
        self.f = torch.sin

        self.reset_parameters()

    def reset_parameters(self):
        initrange = 1.0 / math.sqrt(self.opt_dim)

        for weight in self.parameters():
            weight.data.uniform_(-initrange, initrange)

    def forward(self, x):
        v1 = self.f(torch.matmul(x, self.w) + self.b)
        v2 = torch.matmul(x, self.w0) + self.b0
        vec = torch.cat([v1, v2], -1)
        return vec


class TimeFeatureEmbedding(nn.Module):
    def __init__(self, d_model, embed_type='timeF', freq='h'):
        super(TimeFeatureEmbedding, self).__init__()
        # [s:secondly, t:minutely, h:hourly, d:daily, b:business days, w:weekly, m:monthly]
        d_inp = 30000
        self.embed = nn.Linear(d_inp, d_model, bias=False)

    def forward(self, x):
        return self.embed(x)

class DataEmbedding(nn.Module):
    def __init__(self, c_in, d_model, dropout=0.05, conv_kernel=0, embConstruct=0, use_batchnorm=False, embed_type='timeF', freq='d'):
        super(DataEmbedding, self).__init__()

        self.position_embedding = PositionalEmbedding(d_model=d_model)
        # self.temporal_embedding = TimeFeatureEmbedding(d_model=d_model, embed_type=embed_type, freq=freq)
        self.time_embedding = LinearEmbedding(c_in=3, d_model=d_model)
        self.dropout = nn.Dropout(p=dropout)

        self.value_embedding_linear = LinearEmbedding(c_in=c_in, d_model=d_model)

        if conv_kernel>0:
            # Chaopeng Shen
            if embConstruct==0:
                self.conv_embedding = CausalConv1d(in_channels=c_in, out_channels=d_model, kernel_width=conv_kernel, use_batchnorm=use_batchnorm)
            elif embConstruct==1:
                self.conv_embedding = CausalConv1dAlt(in_channels=c_in, out_channels=d_model, kernel_width=conv_kernel, use_batchnorm=use_batchnorm)
            elif embConstruct==2:
                self.conv_embedding = CausalConv1dAlt2(in_channels=c_in, out_channels=d_model, kernel_width=conv_kernel, use_batchnorm=use_batchnorm)
        else:
            self.conv_embedding = None

    def forward(self, x, x_mark):
        if self.conv_embedding is not None:
            x = self.value_embedding_linear(x) + self.time_embedding(x_mark) + self.position_embedding(x) + self.conv_embedding(x)
        else:
            x = self.value_embedding_linear(x) + self.time_embedding(x_mark) + self.position_embedding(x)

        return self.dropout(x)

