import os
import glob
import pandas as pd
import numpy as np
import warnings
from waterGPT.utils.stats.transform import normalization
from waterGPT.utils.stats.transform import load_stat
from waterGPT.utils.data.data_integration import data_integration

class LoadCSV(object):
    __name__ = "LoadCSV"

    def __init__(self, inp_folder, select_date_list=None, csv_date_list=None, load_all=False):

        try:
            if csv_date_list is None:
                csv_date_list = self.get_all_date_range(inp_folder)
            if load_all or (select_date_list is None):
                select_date_list = csv_date_list
        except:
            pass

        self.inp_folder = inp_folder
        self.select_date_list = select_date_list
        self.csv_date_list = csv_date_list

    def get_all_date_range(self, inp_folder):
        inp_folder_time = glob.glob(os.path.join(inp_folder, "18*")) + glob.glob(os.path.join(inp_folder, "19*")) + \
                          glob.glob(os.path.join(inp_folder, "20*")) + glob.glob(os.path.join(inp_folder, "21*"))

        year_list = [int(os.path.split(x)[-1]) for x in inp_folder_time]
        year_list = np.sort(year_list)
        # for start
        # check all size
        csv_path_list = glob.glob(os.path.join(inp_folder, str(year_list[0]), "*.csv"))
        self.check_shape(csv_path_list)
        df_str = pd.read_csv(csv_path_list[0], header=None)
        num_str = df_str.shape[-1]
        str_date = pd.date_range("{}-12-31".format(year_list[0]), "{}-12-31".format(year_list[0])).shift(-num_str + 1)[
            0]
        str_date = str_date.strftime("%Y-%m-%d")

        df_end = pd.read_csv(glob.glob(os.path.join(inp_folder, str(year_list[-1]), "*.csv"))[0], header=None)
        num_str = df_end.shape[-1]
        end_date = pd.date_range("{}-01-01".format(year_list[-1]), "{}-01-01".format(year_list[-1])).shift(num_str - 1)[
            0]
        end_date = end_date.strftime("%Y-%m-%d")

        return [str_date, end_date]

    def check_shape(self, csv_path_list):

        for idx, csv_path in enumerate(csv_path_list):
            df_shape = pd.read_csv(csv_path).shape
            if idx == 0:
                csv_shape = df_shape
            else:
                assert csv_shape == df_shape

    def remove_nan(self, data, fill_num=0, nan_list=[-9999, 9999]):
        data[data >= nan_list[-1]] = np.nan
        data[data <= nan_list[0]] = np.nan
        data[np.isnan(data)] = fill_num

        return data


    def load_time_series(self, var_list, stat_list=None, remove_nan=True, fill_num=0, nan_list=[-9999, 9999]):
        """
        var_list: ["precipitation", "temperature", ]
        # do normalization
        stat_list = [[90 percentile, 10 percentile, mean, std],  [90 percentile, 10 percentile, mean, std]...]

        """
        if len(var_list) == 0 or var_list is None:
            warnings.warn("loading timeseries is empty or None!")
            return []

        var_type = "var_t"

        for ndx, var_s in enumerate(var_list):

            data_all = self.load_all_csv(var_s, var_type, self.select_date_list, self.csv_date_list)  # pxiel, time

            # normalization
            if not stat_list is None:
                data_all = normalization(data_all, stat_list=[stat_list[var_s]], from_raw=True)

            # remove nan with 0
            if remove_nan:
                data_all = self.remove_nan(data=data_all, fill_num=fill_num, nan_list=nan_list)

            data_all = data_all[:, :, None]  # pixel, time, 1

            if ndx == 0:
                data = data_all
            else:
                data = np.concatenate((data, data_all), axis=2)  # pixel, time, features
            print("{} (remove nan) min : {:.3f} max: {:.3f} median: {:.3F}".format(var_s, np.nanmin(data_all), np.nanmax(data_all), np.nanmedian(data_all)))
        assert len(data.shape) == 3
        return data

    def load_constant(self, var_list, stat_list=None, remove_nan=True, fill_num=0, convert_time_series=True,
                      specific_folder=False, nan_list=[-9999, 9999]):

        """
        var_list: ["precipitation", "temperature", ]
        # do normalization
        stat_list = [[90 percentile, 10 percentile, mean, std],  [90 percentile, 10 percentile, mean, std]...]

        """
        if len(var_list) == 0 or var_list is None:
            warnings.warn("loading constant is empty or None!")
            return []

        var_type = "var_c"

        for ndx, var_s in enumerate(var_list):

            if var_s in ["aspectcosine"]:
                nan_list = [-100, 9999]

            data_all = self.load_all_csv(var_s, var_type, specific_folder=specific_folder)

            # assert np.squeeze(data_all) == 1

            if len(data_all.shape) == 1:
                data_all = data_all[:,None] # pixel, features
            assert len(data_all.shape) == 2

            # normalization
            if not stat_list is None:
                data_all = normalization(data_all, stat_list=[stat_list[var_s]], from_raw=True)

            # remove nan
            if remove_nan:
                data_all = self.remove_nan(data=data_all, fill_num=fill_num, nan_list=nan_list)  # pixel, features

            if ndx == 0:
                data = data_all
            else:
                data = np.concatenate((data, data_all), axis=-1)  # pixel, features

            print("{} (remove nan) min : {:.3f} max: {:.3f} median: {:.3F}".format(var_s, np.nanmin(data_all),
                                                                               np.nanmax(data_all),
                                                                               np.nanmedian(data_all)))

        if convert_time_series:
            data = self.constant_to_time_series(data, self.select_date_list) # pixel, time, features



        return data

    def load_folder(self, var_list):
        return self.load_constant(var_list, do_norm=False, remove_nan=False, convert_time_series=False,
                                  specific_folder=True)

    def load_all_csv(self, var_s, var_type, select_date_list=None, csv_date_list=None, specific_folder=False):

        # load constant data
        if var_type in ["var_c"]:
            if not specific_folder:
                path_cData = os.path.join(self.inp_folder, 'const', var_s + '.csv')
            else:
                path_cData = os.path.join(self.inp_folder, var_s + '.csv')
            df = pd.read_csv(path_cData, dtype=np.float, header=None, )  # pixel,

            df_a = df.values  # pixel,
            data = df_a
            data[data == -9999] = np.nan

            return data

        if var_type in ["var_t"]:

            select_str_s, select_end_s = select_date_list
            select_date_range = pd.date_range(select_str_s, select_end_s)  # e.g. 2015
            select_str_year = int(select_date_range[0].strftime('%Y'))
            select_end_year = int(select_date_range[-1].strftime('%Y'))

            for ndx_year, year_int in enumerate(np.arange(select_str_year, select_end_year + 1)):

                path_xdata = os.path.join(self.inp_folder, str(year_int), var_s + '.csv')
                df = pd.read_csv(path_xdata, dtype=np.float, header=None)  # pixel, time

                df_a = df.values  # pixel, time
                if ndx_year == 0:
                    data = df_a
                else:
                    data = np.concatenate((data, df_a), axis=1)  # fsg pixel, times

            # select data within select_date_range
            csv_str_date_s = csv_date_list[0]
            csv_end_date_s = csv_date_list[-1]

            # create sub time series
            if str(select_str_year) == csv_str_date_s[:4]:
                sub_str_date_s = csv_str_date_s
            else:
                sub_str_date_s = str(select_str_year) + '-01-01'

            if str(select_end_year) == csv_end_date_s[:4]:
                sub_end_date_s = csv_end_date_s
            else:
                sub_end_date_s = str(select_end_year) + '-12-31'

            sub_date_index = pd.date_range(sub_str_date_s, sub_end_date_s)
            ts = pd.Series(np.arange(len(sub_date_index)), index=sub_date_index)
            sub_data = data[:, ts[select_date_range].values]  # pixel, target time
            sub_data[sub_data == -9999] = np.nan
            return sub_data

    def constant_to_time_series(self, data, date_list):
        # data shape: pixel, features
        date_range = pd.date_range(date_list[0], date_list[-1])
        data = data[:, None, :]  # pix, 1, feas
        data = np.repeat(data, len(date_range), axis=1)
        return data

def load_xy(csv_path, select_date_list, var_x_list=None, var_c_list=None, var_y_list=None, config_dict=None, stat_dict=None, DI=None, csv_date_range=None,not_load_y=False):
    """
    DI: data integration
    output size: (basin, time, features)
    """
    print("loading data ...")

    if var_x_list is None:
        var_x_list = config_dict["var_times"]
    if var_c_list is None:
        var_c_list = config_dict["var_const"]
    if var_y_list is None:
        var_y_list = config_dict["target"]

    if stat_dict is None:
        stat_dict = load_stat(csv_folder=csv_path, var_x_list=var_x_list, var_c_list=var_c_list,
                                                      var_y_list=var_y_list)
    train_load_csv = LoadCSV(inp_folder=csv_path, select_date_list=select_date_list, csv_date_list=csv_date_range)
    x = train_load_csv.load_time_series(var_x_list, stat_list=stat_dict["x"], remove_nan=True, fill_num=0)
    c = train_load_csv.load_constant(var_c_list, stat_list=stat_dict["c"], remove_nan=True, fill_num=0, convert_time_series=True)
    x = np.concatenate([x, c], axis=-1)
    if (not var_y_list is None) and (not not_load_y):
        y = train_load_csv.load_time_series(var_y_list, stat_list=stat_dict["y"], remove_nan=False)
    else:
        y = None

    if not DI is None:
        x, y = data_integration(x,y, DI=DI, time_dim=1)
    try:
        print("done!")
        # DON NOT USE NANMIN, NANMAX, NANMEDIAN
        print("x (remove nan) min : {:.3f} max: {:.3f} median: {:.3F}".format(x.min(), x.max(), np.median(x)))
        # USE NANMIN, NANMAX, NANMEDIAN
        print("y (keep   nan) min : {:.3f} max: {:.3f} median: {:.3F}".format(np.nanmin(y), np.nanmax(y), np.nanmedian(y)))
    except:
        pass

    return x, y

