import os
import json
import glob
import shutil
import numpy as np
import pandas as pd
import random
import torch
import inspect
from matplotlib import pyplot as plt
import warnings

from waterGPT.utils.sys_tools import re_folder
from waterGPT.utils.stats.transform import align_dim
from waterGPT.utils.stats.transform import de_normalization
from waterGPT.utils.data.data_integration import shift_date


class SaveInfo(object):

    def __init__(self, save_path=None):

        if save_path is None:
            save_path = "./"

        self.save_path = save_path

    def seed(self, seed):
        np.random.seed(seed)
        random.seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        os.environ["PYTHONHASHSEED"] = str(seed)

    def update_config(self, *args, save_path=None, rename_output=True, **kwargs):
        config_dict = self.save_config_info(*args, save_path=save_path, rename_output=rename_output, **kwargs)
        return config_dict

    def save_config_info(self, *args, save_path=None, rename_output=True, **kwargs):

        """
        e.g.
        para1 = "a"
        para2 = "b"
        save_config_info(para1=para1, para2=para2)
        """

        if rename_output:
            # rename output
            output_new = self.create_save_folder(**kwargs)
            kwargs["output_path"] = output_new
        else:
            output_new = kwargs["output_path"]

        # update default save path
        self.update(save_path=output_new)

        if save_path is None:
            save_path = self.save_path

        re_folder(save_path)

        save_file = os.path.join(save_path, "config_info.json")
        if len(args):
            config_dict = {("arg" + str(idx + 1)): arg for idx, arg in enumerate(args)}
        else:
            config_dict = {}

        config_dict.update(**kwargs)

        with open(save_file, 'w') as fout:
            json.dump(config_dict, fout, indent=4)

        # keep seed
        self.seed(kwargs["SEED"])

        return config_dict

    def save_model(self, model, epoch=None, save_name="saved_model.pt", save_path=None, optim=None, accelerator=None):

        if save_path is None:
            save_path = self.save_path
            save_path = os.path.join(save_path, "model")

        re_folder(save_path)

        if not epoch is None:
            save_name = "model_Ep{}.pt".format(epoch)

        model_file = os.path.join(save_path, save_name)
        if accelerator is None:
            torch.save(model, model_file)

            # save model parameters
            model_file_state = model_file.replace(".pt", ".state")
            state = {'model_state': model.state_dict(),
                     'model_name': type(model).__name__,
                     # 'optimizer_state': optim.state_dict(),
                     # 'optimizer_name': type(optim).__name__,
                     'epoch': epoch,
                     }
            torch.save(state, model_file_state)
        else:
            model = accelerator.unwrap_model(model)
            # accelerator.wait_for_everyone()
            accelerator.save(model, model_file.replace(".pt", ""))

    def save_best_model(self, model, valid_logs, epoch, best_score, score_class="Corr", save_path=None,
                        save_each=False, optim=None, nth_save_model=False, **kwargs):
        try:
            accelerator = kwargs["accelerator"]
        except:
            accelerator = None

        if save_path is None:
            save_path = self.save_path
            save_path = os.path.join(save_path, "model")
            save_path_metric = os.path.join(save_path, "..",  "metrics")

        # save each epoch
        if save_each:
            self.save_model(model, epoch=epoch, optim=optim, accelerator=accelerator)
            print("Model saved!")

        # save nth epoch
        if nth_save_model:
            if (epoch + 1) % nth_save_model == 0:
                self.save_model(model, epoch=epoch, optim=optim, accelerator=accelerator)
                print("Model saved!")

        if not valid_logs is None:
            best_metric_file = os.path.join(save_path_metric, "best_metrics.json")

            if best_score < valid_logs[score_class]:
                best_score = valid_logs[score_class]
                self.save_model(model, save_name="best_model.pt", optim=optim, accelerator=accelerator)
                print("Best model saved!")
                valid_logs["epoch"] = epoch
                with open(best_metric_file, 'w') as fout:
                    json.dump(valid_logs, fout, indent=4)

        return best_score

    def load_model(self, model_file=None, epoch=None, save_name="saved_model.pt", save_path=None, **kwargs):

        try:
            accelerator = kwargs["accelerator"]
        except:
            accelerator = None

        if model_file is None:
            if save_path is None:
                save_path = self.save_path
                save_path = os.path.join(save_path, "model")

            if not epoch is None:
                save_name = "model_Ep{}.pt".format(epoch)
            model_file = os.path.join(save_path, save_name)

        if accelerator is None:
            try:
                model = torch.load(model_file)
            except:
                model = torch.load(model_file,  map_location='cpu')
                model = model.cuda()
        else:
            model = accelerator.load_state(model_file)

        return model

    def create_save_folder(self, **kwargs):
        """
        e.g.
        output = "/data/output/experiment/"

        new output = "/data/output/experiment/EPOCH_050_RHO_256_BATCHSIZE_128_HIDDENSIZE_256_DROPOUT_0.1_train_2017040120220331_valid_2015040120170331"
        or
        new output = "/data/output/experiment/EPOCH_050_RHO_256_BATCHSIZE_128_HIDDENSIZE_256_DROPOUT_0.1_train_2017040120220331_valid_2015040120170331/kfold_0, kfold_1"
        """

        output_path = kwargs["output_path"]
        EPOCH = kwargs["EPOCH"]
        RHO = kwargs["RHO"]
        BATCH_SIZE = kwargs["BATCH_SIZE"]
        HIDDENSIZE = kwargs["HIDDENSIZE"]
        train_date_list = kwargs["train_date_list"]
        valid_date_list = kwargs["valid_date_list"]
        DROPOUT = kwargs["DROPOUT"]
        nth_kfold = kwargs["nth_kfold"]
        LR = kwargs["LR"]
        DI = kwargs["DI"]

        suffix_DI = ""
        if not DI is None:
            suffix_DI = suffix_DI + "_DI_{:03d}".format(DI)

        # # parameters for transformer
        # Q = kwargs["Q"]
        # V = kwargs["V"]
        H = kwargs["H"]
        N_LAYERS = kwargs["N_LAYERS"]
        # ATTENTION_SIZE = kwargs["ATTENTION_SIZE"]
        dim_feedforward = kwargs["dim_feedforward"]

        # if any([Q, V, H, N_LAYERS, ATTENTION_SIZE]):
        #     for x in [Q, V, H, N_LAYERS, ATTENTION_SIZE]:
        if any([H, N_LAYERS]):
            for x in [H, N_LAYERS]:
                assert x is not None, "Please specify the parameters for transformer!"
            suffix_transformer = "_H{:03d}_N{:03d}_HIDDEN_{}_LR{}".format(H, N_LAYERS, dim_feedforward, LR)
        else:
            suffix_transformer = ""

        # format date
        train_date_list = pd.to_datetime(train_date_list).strftime("%Y%m%d")
        train_date_list = "_".join(list(train_date_list))
        valid_date_list = pd.to_datetime(valid_date_list).strftime("%Y%m%d")
        valid_date_list = "_".join(list(valid_date_list))

        tmp_str = "EPOCH_{:03d}_RHO_{:03d}_BATCHSIZE_{:03d}_HIDDENSIZE_{:03d}_DROPOUT_{}_train_{}_valid_{}{}{}".format(
            EPOCH, RHO, BATCH_SIZE, HIDDENSIZE, DROPOUT, train_date_list, valid_date_list, suffix_DI, suffix_transformer)
        output_path_new = os.path.join(output_path, tmp_str)

        if not nth_kfold is None:
            output_path_new = os.path.join(output_path_new, "kfold_{}".format(nth_kfold))

        return output_path_new

    def save_metrics(self, logs_dict, epoch, stage_name, save_path=None, save_name="log.csv"):

        """
        Continuous write file

        logs_dict = {"loss":0.73, "Corr":0.84"}
        epoch = 0
        stage_name = "Train"

        Syntax:
            save_metrics(logs_dict, epoch, stage_name)
        """
        assert stage_name == "train" or stage_name == "valid"
        logs_dict["epoch"] = epoch
        logs_dict["stage_name"] = stage_name

        if save_path is None:
            save_path = self.save_path
            save_path = os.path.join(save_path, "metrics")

        re_folder(save_path)

        save_file = os.path.join(save_path, save_name)

        logs_list = []
        [logs_list.extend([k, str(v)]) for k, v in logs_dict.items()]
        logs_str = ",".join(logs_list) + "\n"

        # logs_df = pd.DataFrame([logs_list])
        # logs_df.to_csv(save_file, header=None, index=None, mode="a")

        with open(save_file, "a") as out_file:
            out_file.write(logs_str)

    def save_pred_data(self, pred_data, date_list, var_list, stat_list, save_path=None, pred_folder_name="prediction", data_range=None,
                       eps=10e-6, logs=None, logs_name="logs.csv", obs_data=None, DI=None):
        """
        pred_data and obs data have been de-normalized
        pred_data: pixels, time, features
        date_list = [str_date, end_date]
        var_list = ["soil_moisture"]
        save_path = "/data/format_data/"

        "save_path/output_path" is the top path in the output. "pred_folder_name" is a subpath of "save_path" to store pred CSV files.

        -----------------------------------
        --output_path
        --    metrics
        --        Corr_np.jpg
        --        log.csv
        --    model
        --        best_model.pt5
        --        best_metrics.json
        --    pred_folder_name
        --        2021
        --        2022
        --    config_info.json
        --    test_site_metrics.json
        -----------------------------------

        Syntax:
            save_pred_data(pred_data, date_list, band_list)
        """

        if save_path is None:
            save_path = self.save_path
        pred_path = os.path.join(save_path, pred_folder_name)
        re_folder(pred_path)

        if not logs is None:
            save_logs = os.path.join(save_path, logs_name)
            file_suffix = os.path.splitext(logs_name)[-1]
            if file_suffix in [".csv"]:

                if type(logs) == dict:
                    logs_df = pd.DataFrame([logs])
                else:
                    logs_df = logs
                logs_df.to_csv(save_logs)
                # else:
                #     with open(save_logs, "a") as out_file:
                #         out_file.write(logs + "\n")

            elif file_suffix in [".json"]:
                with open(save_logs, "w") as fout:
                    json.dump(logs, fout, indent=4)
            else:
                raise ValueError

        # keep dimension to 3
        data = align_dim(pred_data, dim=3)

        # remove extreme data
        data = data.astype(float)
        if not data_range is None:
            data_max = max(data_range) + eps
            data_min = min(data_range) - eps
            data[data >= data_max] = np.nan
            data[data <= data_min] = np.nan
        data[data < -99] = np.nan
        data[np.isnan(data)] = -9999

        if (not obs_data is None) and (not len(obs_data) == 0):
            obs_data = align_dim(obs_data, dim=3)
            obs_data = de_normalization(obs_data, stat_list=stat_list)
            np.save(os.path.join(pred_path, "pred.npy"), pred_data)
            np.save(os.path.join(pred_path, "obs.npy"), obs_data)

        if not DI is None:
            str_date, end_date = shift_date(DI=DI, date_list=date_list)
        else:
            str_date, end_date = date_list

        # create date range
        ts, year_list = self.create_ts(str_date=str_date, end_date=end_date, freq="D")
        # save data for each year
        for year in year_list:
            select_ts = ts[str(year)]
            date_index = select_ts.values
            data_sel = data[:, date_index, :]
            assert len(data_sel.shape) == 3

            # save datasets
            csv_path = os.path.join(pred_path, str(year))
            re_folder(csv_path)

            for idx_var, band_s in enumerate(var_list):
                data_sel_var = data_sel[:, :, idx_var]  # pixels, time
                df = pd.DataFrame(data_sel_var)
                df.to_csv(os.path.join(csv_path, band_s + ".csv"), header=None, index=None)
                print("Saved! Prediction: {} {}".format(year, band_s))

    def create_ts(self, str_date, end_date, freq="D"):
        date_range = pd.date_range(str_date, end_date, freq=freq)
        year_list = np.sort(date_range.year.unique().to_list())
        ts = pd.Series(index=date_range, data=np.arange(len(date_range)))
        return ts, year_list

    def save_plot(self, save_path=None, save_name="log.csv"):
        if save_path is None:
            save_path = self.save_path
            save_path = os.path.join(save_path, "metrics")

        re_folder(save_path)

        save_file = os.path.join(save_path, save_name)
        logs_df = pd.read_csv(save_file, header=None)
        df_value = logs_df.loc[:, 1::2]
        df_name = logs_df.loc[:, ::2].loc[0, :].values.tolist()
        df_value.columns = df_name

        df_train = df_value[df_value["stage_name"] == "train"]
        df_valid = df_value[df_value["stage_name"] == "valid"]

        # DO NOT ASCENDING!

        # start from epoch 0
        train_zero = np.where(df_train["epoch"].values == 0)[0][-1]
        valid_zero = np.where(df_valid["epoch"].values == 0)[0][-1]

        # drop out stage_name and epoch
        df_train = df_train.drop(['stage_name', 'epoch'], axis=1, inplace=False)
        df_valid = df_valid.drop(['stage_name', 'epoch'], axis=1, inplace=False)

        metric_list = df_train.columns
        train_values = df_train.values[train_zero:, :]
        valid_values = df_valid.values[valid_zero:, :]

        for ndx, metric in enumerate(metric_list):
            self.plot_curve(metric, train_values[:, ndx], valid_values[:, ndx], save_path)

    def plot_curve(self, name, train_data, valid_data, save_path=None):

        if len(train_data) - len(valid_data) == 1:
            train_data = train_data[:-1]

        if save_path is None:
            save_path = self.save_path
            save_path = os.path.join(save_path, "metrics")

        re_folder(save_path)

        epoch_list = range(len(train_data))

        save_file = os.path.join(save_path, name + ".jpg")

        fig, ax = plt.subplots()

        ax.plot(epoch_list, train_data, label='training')
        ax.plot(epoch_list, valid_data, label='validation', linestyle="--")

        # set y axis limit
        ax.set_ylim([0, 1])

        ax.set_xlabel('Epoch', fontsize=16)
        ax.set_ylabel(name, fontsize=16)
        ax.legend(fontsize=14, frameon=False)

        plt.tick_params(axis='both', which='major', labelsize=14)
        plt.title(name, fontsize=16)

        plt.savefig(save_file,
                    dpi=600,
                    orientation='landscape',
                    bbox_inches='tight', pad_inches=0
                    )
        plt.close()

    def check_output_files(self, save_path=None):

        # not debugging
        if self.detect_debugger(): return None

        if save_path is None:
            save_path = self.save_path

        config_info_file = os.path.join(save_path, "config_info.json")
        file_list = glob.glob(os.path.join(save_path, "*"))
        file_list.remove(config_info_file)
        path_exists_bool = len(file_list) > 0

        yea_list = ['yes', 'y', 'ye', '']

        if path_exists_bool:
            choice = input("The output folder already exists, do you want to overwrite it [y/n]?").lower()
            if choice in yea_list:
                choice = input("Do you want to keep the these files?  [y/n]?").lower()
                if not choice in yea_list:
                    config_dict = json.load(open(config_info_file, 'r'))

                    shutil.rmtree(save_path)
                    re_folder(save_path)

                    with open(config_info_file, 'w') as fout:
                        json.dump(config_dict, fout, indent=4)

            else:
                print("Please change your output folder!")
                raise IOError

    def detect_debugger(self):
        """
        detects if debugging is applied
        """

        for frame in inspect.stack():
            if frame[1].endswith("pydevd.py"):
                return True
        return False

    # update parameters
    def update(self, **kwargs):
        '''
        only update one param! don't cal with param in the class!!
        the number of k,v,h according to the param you need to update
        '''
        for k, v in kwargs.items():
            if not hasattr(self, k):
                warnings.warn("There is no argument: {} !".format(k))
            setattr(self, k, v)  # set param according to dict's key
