from . import base
from . import functional as F
from .base import Activation
from waterGPT.utils.math.torch_numpy import convert_to_numpy
import torch
from waterGPT.utils.math.torch_numpy import nanmedian
import numpy as np
import scipy.stats
from waterGPT.utils.stats.transform import align_dim

class Accuracy(base.Metric):
    def __init__(self, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, pred, obs):
        pred = self.activation(pred)
        return F.accuracy(
            pred,
            obs,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Bias_np(base.Metric):
    __name__ = "Bias_np"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs):
        pred = convert_to_numpy(pred)
        obs = convert_to_numpy(obs)

        pred = align_dim(pred, dim=2)
        obs = align_dim(obs, dim=2)

        stat_dict = statError(pred, obs)
        Bias = stat_dict["Bias"]
        Bias = np.nanmedian(Bias).astype(np.float32)
        return Bias


class Bias(base.Metric):
    __name__ = "Bias"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs, time_dim=1):
        assert pred.shape == obs.shape

        x, y = pred, obs

        bias = torch.nanmean(x - y, axis=time_dim)

        bias = nanmedian(bias)
        return bias


class Corr_np(base.Metric):
    __name__ = "Corr_np"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs):
        pred = convert_to_numpy(pred)
        obs = convert_to_numpy(obs)

        pred = align_dim(pred, dim=2)
        obs = align_dim(obs, dim=2)

        # input size: grid, time
        ngrid, nt = pred.shape
        Corr = np.full(ngrid, np.nan)
        for k in range(0, ngrid):
            x = pred[k, :]
            y = obs[k, :]
            ind = np.where(np.logical_and(~np.isnan(x), ~np.isnan(y)))[0]
            if ind.shape[0] > 0:
                xx = x[ind]
                yy = y[ind]
                Corr[k] = scipy.stats.pearsonr(xx, yy)[0]

        Corr = np.float(np.nanmedian(Corr))
        # Corr = np.nanmedian(Corr).astype(np.float32)
        return Corr


class Corr(base.Metric):
    __name__ = "Corr"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs, time_dim=1):
        assert pred.shape == obs.shape
        dim = time_dim

        x, y = pred, obs

        # index_nan = torch.logical_or(torch.isnan(x), torch.isnan(y))
        index_nan = ~(x == y)
        x[index_nan] = torch.nan
        y[index_nan] = torch.nan

        x_mean = torch.nanmean(x, keepdims=True, dim=dim)
        y_mean = torch.nanmean(y, keepdims=True, dim=dim)

        centered_x = x - x_mean
        centered_y = y - y_mean

        covariance = (centered_x * centered_y)
        covariance = torch.nansum(covariance, keepdims=True, dim=dim)

        # bessel_corrected_covariance = covariance

        tmp_x = torch.nansum(centered_x ** 2, keepdims=True, dim=dim)
        tmp_y = torch.nansum(centered_y ** 2, keepdims=True, dim=dim)

        corr = covariance / torch.sqrt(tmp_x * tmp_y)
        corr = torch.flatten(corr)
        corr = torch.nanquantile(corr, q=0.5)
        corr = torch.nanmean(corr)
        return corr


class Fscore(base.Metric):
    def __init__(self, beta=1, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.beta = beta
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, pred, obs):
        pred = self.activation(pred)
        return F.f_score(
            pred,
            obs,
            eps=self.eps,
            beta=self.beta,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class IoU(base.Metric):
    __name__ = "iou_score"

    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, pred, obs):
        pred = self.activation(pred)
        return F.iou(
            pred,
            obs,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class NSE_np(base.Metric):
    __name__ = "NSE_np"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs):
        pred = convert_to_numpy(pred)
        obs = convert_to_numpy(obs)

        pred = align_dim(pred, dim=2)
        obs = align_dim(obs, dim=2)

        stat_dict = statError(pred, obs)
        NSE = stat_dict["NSE"]
        NSE = np.nanmedian(NSE).astype(np.float32)
        return NSE


class NSE(base.Metric):
    __name__ = "NSE"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs, time_dim=1):
        assert pred.shape == obs.shape

        x, y = pred, obs

        index_nan = torch.logical_or(torch.isnan(pred), torch.isnan(obs))
        pred[index_nan] = torch.nan
        obs[index_nan] = torch.nan

        obs_mean = torch.nanmean(obs, dim=time_dim, keepdim=True)

        SST = torch.nansum((obs - obs_mean) ** 2, dim=time_dim, keepdim=True)
        SSRes = torch.nansum((obs - pred) ** 2, dim=time_dim, keepdim=True)

        nse = 1 - SSRes / SST
        nse[torch.logical_or(nse == torch.inf, nse == -torch.inf)] = torch.nan
        nse = nanmedian(nse)

        return nse


class Precision(base.Metric):
    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, pred, obs):
        pred = self.activation(pred)
        return F.precision(
            pred,
            obs,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class Recall(base.Metric):
    def __init__(self, eps=1e-7, threshold=0.5, activation=None, ignore_channels=None, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        self.threshold = threshold
        self.activation = Activation(activation)
        self.ignore_channels = ignore_channels

    def forward(self, pred, obs):
        pred = self.activation(pred)
        return F.recall(
            pred,
            obs,
            eps=self.eps,
            threshold=self.threshold,
            ignore_channels=self.ignore_channels,
        )


class RMSE_np(base.Metric):
    __name__ = "RMSE_np"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs):
        pred = convert_to_numpy(pred)
        obs = convert_to_numpy(obs)

        pred = align_dim(pred, dim=2)
        obs = align_dim(obs, dim=2)

        stat_dict = statError(pred, obs)
        RMSE = stat_dict["RMSE"]
        RMSE = np.nanmedian(RMSE).astype(np.float32)
        return RMSE


class RMSE(base.Metric):
    __name__ = "RMSE"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs, time_dim=1):
        assert pred.shape == obs.shape

        x, y = pred, obs

        rmse = torch.sqrt(torch.nanmean((x - y) ** 2, axis=time_dim))

        rmse = nanmedian(rmse)
        return rmse


class ubRMSE_np(base.Metric):
    __name__ = "ubRMSE_np"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs):
        pred = convert_to_numpy(pred)
        obs = convert_to_numpy(obs)

        pred = align_dim(pred, dim=2)
        obs = align_dim(obs, dim=2)

        stat_dict = statError(pred, obs)
        ubRMSE = stat_dict["ubRMSE"]
        ubRMSE = np.nanmedian(ubRMSE).astype(np.float32)
        return ubRMSE


class ubRMSE(base.Metric):
    __name__ = "ubRMSE"

    def __init__(self, ):
        super().__init__()

    def forward(self, pred, obs, time_dim=1):
        assert pred.shape == obs.shape

        x, y = pred, obs

        index_nan = torch.logical_or(torch.isnan(x), torch.isnan(y))
        x[index_nan] = torch.nan
        y[index_nan] = torch.nan

        x_mean = torch.nanmean(x, axis=time_dim, keepdim=True)
        y_mean = torch.nanmean(y, axis=time_dim, keepdim=True)

        x_anom = x - x_mean
        y_anom = y - y_mean

        ubRMSE = torch.sqrt(torch.nanmean((x_anom - y_anom) ** 2, axis=time_dim, keepdim=True))

        ubRMSE = nanmedian(ubRMSE)
        return ubRMSE

keyLst = ['Bias', 'RMSE', 'NSE', 'Corr']

def statError(pred, target):
    import warnings
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=RuntimeWarning)
        ngrid, nt = pred.shape
        # Bias
        Bias = np.nanmean(pred - target, axis=1)
        # RMSE
        RMSE = np.sqrt(np.nanmean((pred - target)**2, axis=1))
        # ubRMSE
        predMean = np.tile(np.nanmean(pred, axis=1), (nt, 1)).transpose()
        targetMean = np.tile(np.nanmean(target, axis=1), (nt, 1)).transpose()
        predAnom = pred - predMean
        targetAnom = target - targetMean
        ubRMSE = np.sqrt(np.nanmean((predAnom - targetAnom)**2, axis=1))

        # # FDC metric
        # predFDC = calFDC(pred)
        # targetFDC = calFDC(target)
        # FDCRMSE = np.sqrt(np.nanmean((predFDC - targetFDC) ** 2, axis=1))

        # rho R2 NSE
        Corr = np.full(ngrid, np.nan)
        CorrSp = np.full(ngrid, np.nan)
        R2 = np.full(ngrid, np.nan)
        NSE = np.full(ngrid, np.nan)
        PBiaslow = np.full(ngrid, np.nan)
        PBiashigh = np.full(ngrid, np.nan)
        PBias = np.full(ngrid, np.nan)
        PBiasother = np.full(ngrid, np.nan)
        KGE = np.full(ngrid, np.nan)
        KGE12 = np.full(ngrid, np.nan)
        RMSElow = np.full(ngrid, np.nan)
        RMSEhigh = np.full(ngrid, np.nan)
        RMSEother = np.full(ngrid, np.nan)
        for k in range(0, ngrid):
            x = pred[k, :]
            y = target[k, :]
            ind = np.where(np.logical_and(~np.isnan(x), ~np.isnan(y)))[0]
            if ind.shape[0] > 0:
                xx = x[ind]
                yy = y[ind]
                # percent bias
                PBias[k] = np.sum(xx - yy) / np.sum(yy) * 100

                # FHV the peak flows bias 2%
                # FLV the low flows bias bottom 30%, log space
                pred_sort = np.sort(xx)
                target_sort = np.sort(yy)
                indexlow = round(0.3 * len(pred_sort))
                indexhigh = round(0.98 * len(pred_sort))
                lowpred = pred_sort[:indexlow]
                highpred = pred_sort[indexhigh:]
                otherpred = pred_sort[indexlow:indexhigh]
                lowtarget = target_sort[:indexlow]
                hightarget = target_sort[indexhigh:]
                othertarget = target_sort[indexlow:indexhigh]
                PBiaslow[k] = np.sum(lowpred - lowtarget) / np.sum(lowtarget) * 100
                PBiashigh[k] = np.sum(highpred - hightarget) / np.sum(hightarget) * 100
                PBiasother[k] = np.sum(otherpred - othertarget) / np.sum(othertarget) * 100
                RMSElow[k] = np.sqrt(np.nanmean((lowpred - lowtarget)**2))
                RMSEhigh[k] = np.sqrt(np.nanmean((highpred - hightarget)**2))
                RMSEother[k] = np.sqrt(np.nanmean((otherpred - othertarget)**2))

                if ind.shape[0] > 1:
                    # Theoretically at least two points for correlation
                    Corr[k] = scipy.stats.pearsonr(xx, yy)[0]
                    CorrSp[k] = scipy.stats.spearmanr(xx, yy)[0]
                    yymean = yy.mean()
                    yystd = np.std(yy)
                    xxmean = xx.mean()
                    xxstd = np.std(xx)
                    KGE[k] = 1 - np.sqrt((Corr[k]-1)**2 + (xxstd/yystd-1)**2 + (xxmean/yymean-1)**2)
                    KGE12[k] = 1 - np.sqrt((Corr[k] - 1) ** 2 + ((xxstd*yymean)/ (yystd*xxmean) - 1) ** 2 + (xxmean / yymean - 1) ** 2)
                    SST = np.sum((yy-yymean)**2)
                    SSReg = np.sum((xx-yymean)**2)
                    SSRes = np.sum((yy-xx)**2)
                    R2[k] = 1-SSRes/SST
                    NSE[k] = 1-SSRes/SST

        outDict = dict(Bias=Bias, RMSE=RMSE, ubRMSE=ubRMSE, Corr=Corr, CorrSp=CorrSp, R2=R2, NSE=NSE,
                       FLV=PBiaslow, FHV=PBiashigh, PBias=PBias, PBiasother=PBiasother, KGE=KGE, KGE12=KGE12,
                       lowRMSE=RMSElow, highRMSE=RMSEhigh, midRMSE=RMSEother) # fdcRMSE=FDCRMSE,
    return outDict
