import os
import random
import shutil
import numpy as np
import pandas as pd
from distutils.dir_util import copy_tree


def re_folder_rec(path_s):
    b = os.path.normpath(path_s).split(os.sep)  # e.g. ['da','jxl','wp']
    b = [x + "/" for x in b]  # e.g. ['da/','jxl/','wp/']
    fn_rec = [
        "".join(b[0: x[0]]) for x in list(enumerate(b, 1))
    ]  # e.g. ['da/','da/jxl/','da/jxl/wp/']
    fn_None = [os.mkdir(x) for x in fn_rec if not os.path.exists(x)]


def re_folder(path_s, del_old_path=False):
    """
    delete old folder and recreate new one
        Use "try" to avoid errors. When kfold runs, it will create the n process and sometimes it will generate the folder
    even if this folder exists.

    Parameters
        path_s: str, path to folder
        del_old_path: bool, delete old folder or not

    Returns
        None

    Syntax
        re_folder("/mnthh/sdbhh/consthh/demo_data", del_old_path=True)

        or using makedirs(path_s, exist_ok=True)
    """
    if os.path.exists(path_s):
        if del_old_path:
            try:
                shutil.rmtree(path_s)
            except:
                pass
            try:
                re_folder_rec(path_s)
            except:
                pass
        else:
            pass
    else:
        try:
            re_folder_rec(path_s)
        except:
            pass


def mkdir(path_s, del_old_path=False):
    re_folder(path_s=path_s, del_old_path=del_old_path)


def cp(src, dst, forced=True):
    """
    Copy all the files in directory A to directory B

    e.g.
    src: /data/file1.py
              /file2.py
              .empty
    dst: /mnt/sdb/

    cp("/data", "/mnt/sdb")

    output:
        /mnt/sdb/file1.py
                /file2.py
                .empty
    """
    copy_tree(src=src, dst=dst, )


def check_folder_within_dir(inp_path, target_folder, max_depth=1):
    """
    |--mnt
    |   |--data
    |   |   |--forcing
    |   |   |   |--2010
    |   |   |   |--2011
    |   |   |--const
    |   |   |   |--soil.csv
    ...

    inp_path = "/mnt"
    target_folder = "const"

    syntax:
        check_folder_in_dir(inp_path, target_foldr)
        output: False
        check_folder_in_dir(inp_path, target_foldr, max_depth=2)
        output: True
    """

    abs_path = os.path.abspath(inp_path)
    for root, dirs, files in os.walk(abs_path):
        if root[len(abs_path):].count(os.sep) < max_depth:
            if target_folder in dirs:
                return True
        else:
            break
    return False


def check_folder_on_dir(inp_path, target_folder):
    """
    inp_path = "/mnt/sdb/const"
    target_folder = "const"
    check_folder_on_dir(inp_path, target_folder)
    output: True
    """
    inp_path = inp_path.replace("\\", "/")
    inp_path = os.path.normpath(inp_path)
    path_list = inp_path.split(os.sep)
    if target_folder in path_list:
        return True
    else:
        return False

def convert_to_list(data):
    if type(data) == list:
        return data
    if type(data) == np.ndarray:
        return data.tolist()
    if type(data) == pd.DataFrame:
        return data.values.tolist()
    if type(data) == pd.Series:
        return data.values.tolist()
    if type(data) == dict:
        return list(data.items())
    if type(data) in [int, float]:
        return list(data)


def get_key_by_value(data_dict, search_value):
    """
    data_dict = {'a': [16, 17], 'b': [16, 19, 20, 21]}
    search_value = 16

    Syntax:
        get_key_by_value(data_dict, 16)
            output: ["a","b"]
        get_key_by_value(data_dict, 17)
            output: ["a"]
    """

    dict_key = list(data_dict.keys())
    dict_value = list(data_dict.values())
    dict_index = []
    for idx, val in enumerate(dict_value):
        if search_value in val:
            dict_index.append(idx)

    key_list = np.array(dict_key)[dict_index].tolist()
    return key_list

def split_path(inp_file):
    """
    inp_file = "/data/average/basin.shp"

    output,
    inp_path = "/data/average"
    basename = "basin"
    suffix = ".shp"
    """
    inp_path, file_name = os.path.split(inp_file)
    basename, suffix = os.path.splitext(file_name)
    return inp_path, basename, suffix

def mapping_data(data, mapping_key=None, mapping_value=None, mapping_dict=None, outside_use_nan=False, data_type=np.float32, nan_value=None):
    """
    replace matrix values with mapping table

    Parameters
        data: numpy array, any shape
        mapping_key: key of mapping table
        mapping_value: value of mapping table

        mapping_dict: dictionary of mapping table
        outside_use_nan: if True, use nan for outside values

    Returns
        numpy array, same shape as data

    Syntax
        data = np.arange(12.).reshape([3,4])
        mapping_dict = {0: 100, 1: 101, 2: 102, 3: 103, 4: 104, 5: 105, 6: 106, 7: 107, 8: 108}
        mapping_data(data, mapping_dict=mapping_dict, outside_use_nan=True)
    """

    if not mapping_dict is None:
        mapping_key = np.array(list(mapping_dict.keys()))
        mapping_value = np.array(list(mapping_dict.values()))

    # sort mapping key
    index_key = mapping_key.argsort()
    mapping_key = mapping_key[index_key]
    mapping_value = mapping_value[index_key]

    data_shape = data.shape
    data_flatten = data.flatten()
    data_flatten = data_flatten.astype(data_type)

    # generate mask for elements that need replacement
    mask = np.in1d(data, mapping_key)
    # replace elements that in mapping_key
    data_flatten[mask] = mapping_value[np.searchsorted(mapping_key, data_flatten[mask])]

    if outside_use_nan:
        # replace elements that not in mapping_key with nan
        data_flatten[~mask] = np.nan
    else:
        assert not np.sum(~mask), "Not all values are in the mapping table (outside_use_nan=False)"

    if not nan_value is None:
        data_flatten[data_flatten == nan_value] = np.nan

    # reshape data
    data_replaced = data_flatten.reshape(data_shape)

    return data_replaced

def fix_seed(seed=42):
    np.random.seed(seed)
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    try:
        import torch
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
    except:
        pass

# if __name__ == '__main__':
#     pass
